from __future__ import annotations

import itertools
from collections import defaultdict
from collections.abc import Iterator
from dataclasses import asdict, dataclass, field
from itertools import islice
from typing import Literal

import pandas as pd
from cognite.client.data_classes import FileMetadata, Sequence
from cognite.client.data_classes._base import CogniteResource
from cognite.client.data_classes.data_modeling.ids import AbstractDataclass
from cognite.client.data_classes.data_modeling.instances import InstanceCore
from deepdiff import DeepDiff  # type: ignore[import-untyped]
from typing_extensions import TypeAlias

from cognite.powerops.resync.models.base.cdf_resources import CDFResource
from cognite.powerops.resync.models.base.resource_type import Resource

Group: TypeAlias = Literal["CDF", "Domain"]


@dataclass
class Change:
    last: Resource
    new: Resource

    @property
    def changed_fields(self) -> str:
        last_dumped = (
            self.last.dump(camel_case=True)
            if isinstance(self.last, (CogniteResource, AbstractDataclass))
            else self.last.model_dump()
        )
        new_dumped = (
            self.new.dump(camel_case=True)
            if isinstance(self.new, (CogniteResource, AbstractDataclass))
            else self.new.model_dump()
        )
        return (
            DeepDiff(last_dumped, new_dumped, ignore_string_case=True)
            .pretty()
            .replace("added to dictionary.", f"will be added to {self.last.external_id}.")
            .replace("removed from dictionary.", f"will be removed from {self.last.external_id}.")
        )

    @property
    def is_changed_content(self) -> bool:
        if isinstance(self.last, (Sequence, FileMetadata)) and isinstance(self.new, (Sequence, FileMetadata)):
            last_hash = self.last.metadata.get(CDFResource.content_key_hash) if self.last.metadata else None
            new_hash = self.new.metadata.get(CDFResource.content_key_hash) if self.new.metadata else None
            return last_hash != new_hash or last_hash is None or new_hash is None
        return False


@dataclass
class FieldSummary:
    group: Group
    field_name: str
    added: int
    removed: int
    changed: int
    unchanged: int

    @property
    def total(self) -> int:
        return self.added + self.removed + self.changed + self.unchanged


@dataclass
class FieldIds:
    group: Group
    field_name: str
    added: list[str] = field(default_factory=list)
    removed: list[str] = field(default_factory=list)
    changed: list[str] = field(default_factory=list)
    unchanged: list[str] = field(default_factory=list)

    @property
    def total(self) -> int:
        return len(self.added) + len(self.removed) + len(self.changed) + len(self.unchanged)


@dataclass
class FieldDifference:
    group: Group
    field_name: str
    added: list[Resource] = field(default_factory=list)
    removed: list[Resource] = field(default_factory=list)
    changed: list[Change] = field(default_factory=list)
    unchanged: list[Resource] = field(default_factory=list)

    @property
    def total(self) -> int:
        return len(self.added) + len(self.removed) + len(self.changed) + len(self.unchanged)

    @property
    def removed_ids(self) -> list[str]:
        output = []
        for item in self.removed:
            if isinstance(item, InstanceCore):
                output.append(item.as_id())  # type: ignore[union-attr]
            elif isinstance(item, AbstractDataclass):
                output.append(item)
            elif hasattr(item, "external_id"):
                output.append(item.external_id)
            else:
                raise NotImplementedError(f"Could not find identifier of {item}")
        return output

    def __add__(self, other: FieldDifference) -> FieldDifference:
        if self.field_name != other.field_name or self.group != other.group or not isinstance(other, FieldDifference):
            return NotImplemented
        return FieldDifference(
            group=self.group,
            field_name=self.field_name,
            added=self.added + other.added,
            removed=self.removed + other.removed,
            changed=self.changed + other.changed,
            unchanged=self.unchanged + other.unchanged,
        )

    def as_summary(self) -> FieldSummary:
        return FieldSummary(
            group=self.group,
            field_name=self.field_name,
            added=len(self.added),
            removed=len(self.removed),
            changed=len(self.changed),
            unchanged=len(self.unchanged),
        )

    def as_ids(self, limit: int = -1) -> FieldIds:
        def get_identifier(item: Resource) -> str:
            if hasattr(item, "external_id"):
                return item.external_id or "missing external_id"
            elif hasattr(item, "name"):
                return item.name or "missing name"
            raise NotImplementedError(f"Could not find external_id or name in {item}")

        limit_ = limit if limit > 0 else None
        return FieldIds(
            group=self.group,
            field_name=self.field_name,
            added=[get_identifier(item) for item in islice(self.added, limit_)],
            removed=[get_identifier(item) for item in islice(self.removed, limit_)],
            changed=[get_identifier(item.last) for item in islice(self.changed, limit_)],
            unchanged=[get_identifier(item) for item in islice(self.unchanged, limit_)],
        )

    def set_set_dataset(self, dataset: int) -> None:
        for item in itertools.chain(self.added, self.removed, self.unchanged):
            if hasattr(item, "data_set_id"):
                item.data_set_id = dataset
        for item in self.changed:
            if hasattr(item.last, "data_set_id"):
                item.last.data_set_id = dataset
            if hasattr(item.new, "data_set_id"):
                item.new.data_set_id = dataset


@dataclass
class ModelDifferenceSummary:
    model_name: str
    changes: list[FieldSummary]


@dataclass
class ModelDifference:
    model_name: str
    changes: dict[str, FieldDifference] = field(default_factory=dict)

    def __iter__(self) -> Iterator[FieldDifference]:
        return iter(self.changes.values())

    def __contains__(self, item):  # type: ignore[no-untyped-def]
        return item in self.changes

    def __getitem__(self, item):  # type: ignore[no-untyped-def]
        return self.changes[item]

    def __setitem__(self, key, value):  # type: ignore[no-untyped-def]
        self.changes[key] = value

    def __add__(self, other: ModelDifference) -> ModelDifference:
        if self.model_name != other.model_name or not isinstance(other, ModelDifference):
            return NotImplemented
        changes = {}
        for field_name in set(self.changes.keys()) | set(other.changes.keys()):
            if field_name not in self.changes:
                changes[field_name] = other.changes[field_name]
            elif field_name not in other.changes:
                changes[field_name] = self.changes[field_name]
            else:
                changes[field_name] = self.changes[field_name] + other.changes[field_name]
        return ModelDifference(model_name=self.model_name, changes=changes)

    def has_changes(self, group: str, exclude: set | frozenset = frozenset({"timeseries"})) -> bool:
        return any(
            change.total != len(change.unchanged)
            for change in self.changes.values()
            if change.field_name not in exclude and change.group == group
        )

    def filter_out(self, group: Group | None = None, field_names: set | frozenset = frozenset({"timeseries"})) -> bool:
        """
        Removes the field differences which match the filter criteria.

        Args:
            group: The name of the group to remove
            field_names: The field names to remove

        Returns:
            True if any field differences were removed, False otherwise
        """
        has_removed = False
        for field_name in list(self.changes.keys()):
            if field_name in field_names or self.changes[field_name].group == group:
                del self.changes[field_name]
                has_removed = True
        return has_removed

    def as_summary(self) -> ModelDifferenceSummary:
        return ModelDifferenceSummary(
            model_name=self.model_name,
            changes=[change.as_summary() for _, change in sorted(self.changes.items(), key=lambda k: k[0])],
        )


@dataclass
class ModelDifferences:
    models: list[ModelDifference]

    def append(self, model: ModelDifference) -> None:
        self.models.append(model)

    def has_changes(self, group: str, exclude: set | frozenset = frozenset({"timeseries", "labels"})) -> bool:
        return any(m.has_changes(group, exclude) for m in self.models)

    def as_markdown_detailed(self) -> str:
        report = []
        for model in self.models:
            summaries = [change.as_summary() for change in model]
            table = pd.DataFrame(
                [
                    _exclude_keys(asdict(summary), keys=["group", "name"])
                    for summary in summaries
                    if summary.group == "CDF"
                ],
                index=[summary.field_name for summary in summaries if summary.group == "CDF"],
            )

            ids = [change.as_ids(limit=5) for change in model if change.group == "CDF"]
            added_ids = [f"#### Added {id_.field_name}\n{id_.added}" for id_ in ids if id_.added]
            removed_ids = [f"#### Removed {id_.field_name}\n{id_.removed}" for id_ in ids if id_.removed]

            changed_samples = []
            for change, id_ in zip((c for c in model if c.group == "CDF"), ids):
                if change.changed:
                    new_line = "\n * "
                    sample_line = new_line.join(
                        f"{i}: {c.changed_fields}" for c, i in zip(change.changed[:3], id_.changed[:3])
                    )
                    changed_samples.append(f"#### Changed {change.field_name}\n{sample_line}")
            samples = []
            if added_ids:
                samples.append("\n".join(added_ids))
            if removed_ids:
                samples.append("\n".join(removed_ids))
            if changed_samples:
                samples.append("\n".join(changed_samples))
            if samples:
                samples.insert(0, "### Samples")
            samples = "\n".join(samples)

            report.append(
                f"""## {model.model_name}

{table.T.to_markdown()}

**NOTE** Timeseries are not updated by `resync`

{samples}
    """
            )
        return "\n".join(report)

    def as_markdown_summary(self, no_headers: bool = False, skip_domain: bool = True) -> str:
        by_name = defaultdict(list)
        for model in self.models:
            for change in model:
                if skip_domain and change.group == "Domain":
                    continue
                by_name[change.field_name].append(change)
        report = []
        total_added = 0
        total_removed = 0
        total_changed = 0
        for name, changes in by_name.items():
            added = sum(len(change.added) for change in changes)
            removed = sum(len(change.removed) for change in changes)
            change_count = sum(len(change.changed) for change in changes)
            counts = []
            for count_name, count in zip(("added", "removed", "changed"), (added, removed, change_count)):
                if count > 0:
                    counts.append(f"**{count} {count_name}**")
                else:
                    counts.append(f"{count} {count_name}")
            counts_str = ", ".join(counts)
            report.append(f"* **{name}**: {counts_str}")
            total_added += added
            total_removed += removed
            total_changed += change_count

        list_ = "\n".join(report)
        totals = f"{total_added} added, {total_removed} removed, {total_changed} changed"
        if no_headers:
            top_row = f"**Summary:** {totals}"
        else:
            top_row = f"## Summary: {totals}"
        return f"""{top_row}
{list_ if list_ else 'No changes'}
"""

    def as_github_markdown(self) -> str:
        summary = self.as_markdown_summary()
        detailed = self.as_markdown_detailed()
        return f"""{summary}
<details>
<summary>Details</summary>

{detailed}
</details>
"""


def _exclude_keys(d: dict, keys: list[str]) -> dict:
    return {k: v for k, v in d.items() if k not in keys}



from __future__ import annotations

import datetime
import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field
from pydantic import field_validator, model_validator

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    GraphQLCore,
    ResourcesWrite,
)
from ._shop_case import ShopCase, ShopCaseWrite

if TYPE_CHECKING:
    from ._shop_file import ShopFile, ShopFileGraphQL, ShopFileWrite
    from ._shop_scenario import ShopScenario, ShopScenarioGraphQL, ShopScenarioWrite


__all__ = [
    "BenchmarkingShopCase",
    "BenchmarkingShopCaseWrite",
    "BenchmarkingShopCaseApply",
    "BenchmarkingShopCaseList",
    "BenchmarkingShopCaseWriteList",
    "BenchmarkingShopCaseApplyList",
    "BenchmarkingShopCaseFields",
]

BenchmarkingShopCaseFields = Literal["start_time", "end_time", "delivery_date", "bid_generated"]

_BENCHMARKINGSHOPCASE_PROPERTIES_BY_FIELD = {
    "start_time": "startTime",
    "end_time": "endTime",
    "delivery_date": "deliveryDate",
    "bid_generated": "bidGenerated",
}


class BenchmarkingShopCaseGraphQL(GraphQLCore):
    """This represents the reading version of benchmarking shop case, used
    when data is retrieved from CDF using GraphQL.

    It is used when retrieving data from CDF using GraphQL.

    Args:
        space: The space where the node is located.
        external_id: The external id of the benchmarking shop case.
        data_record: The data record of the benchmarking shop case node.
        scenario: The Shop scenario that was used to produce this result
        start_time: The start time of the case
        end_time: The end time of the case
        shop_files: The list of shop files that are used in a shop run. This encompasses all shop files such as case, module series, cut files etc.
        bid_source: The bid source field.
        delivery_date: The delivery date
        bid_generated: Timestamp of when the bid had been generated
    """

    view_id = dm.ViewId("power_ops_core", "BenchmarkingShopCase", "1")
    scenario: Optional[ShopScenarioGraphQL] = Field(default=None, repr=False)
    start_time: Optional[datetime.datetime] = Field(None, alias="startTime")
    end_time: Optional[datetime.datetime] = Field(None, alias="endTime")
    shop_files: Optional[list[ShopFileGraphQL]] = Field(default=None, repr=False, alias="shopFiles")
    bid_source: Optional[str] = Field(default=None, alias="bidSource")
    delivery_date: Optional[datetime.date] = Field(None, alias="deliveryDate")
    bid_generated: Optional[datetime.datetime] = Field(None, alias="bidGenerated")

    @model_validator(mode="before")
    def parse_data_record(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values
        if "lastUpdatedTime" in values or "createdTime" in values:
            values["dataRecord"] = DataRecordGraphQL(
                created_time=values.pop("createdTime", None),
                last_updated_time=values.pop("lastUpdatedTime", None),
            )
        return values

    @field_validator("scenario", "shop_files", "bid_source", mode="before")
    def parse_graphql(cls, value: Any) -> Any:
        if not isinstance(value, dict):
            return value
        if "items" in value:
            return value["items"]
        return value

    def as_read(self) -> BenchmarkingShopCase:
        """Convert this GraphQL format of benchmarking shop case to the reading format."""
        if self.data_record is None:
            raise ValueError("This object cannot be converted to a read format because it lacks a data record.")
        return BenchmarkingShopCase(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecord(
                version=0,
                last_updated_time=self.data_record.last_updated_time,
                created_time=self.data_record.created_time,
            ),
            scenario=self.scenario.as_read() if isinstance(self.scenario, GraphQLCore) else self.scenario,
            start_time=self.start_time,
            end_time=self.end_time,
            shop_files=[shop_file.as_read() for shop_file in self.shop_files or []],
            bid_source=self.bid_source,
            delivery_date=self.delivery_date,
            bid_generated=self.bid_generated,
        )

    def as_write(self) -> BenchmarkingShopCaseWrite:
        """Convert this GraphQL format of benchmarking shop case to the writing format."""
        return BenchmarkingShopCaseWrite(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=0),
            scenario=self.scenario.as_write() if isinstance(self.scenario, GraphQLCore) else self.scenario,
            start_time=self.start_time,
            end_time=self.end_time,
            shop_files=[shop_file.as_write() for shop_file in self.shop_files or []],
            bid_source=self.bid_source,
            delivery_date=self.delivery_date,
            bid_generated=self.bid_generated,
        )


class BenchmarkingShopCase(ShopCase):
    """This represents the reading version of benchmarking shop case.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the benchmarking shop case.
        data_record: The data record of the benchmarking shop case node.
        scenario: The Shop scenario that was used to produce this result
        start_time: The start time of the case
        end_time: The end time of the case
        shop_files: The list of shop files that are used in a shop run. This encompasses all shop files such as case, module series, cut files etc.
        bid_source: The bid source field.
        delivery_date: The delivery date
        bid_generated: Timestamp of when the bid had been generated
    """

    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "power_ops_types", "BenchmarkingShopCase"
    )
    bid_source: Union[str, dm.NodeId, None] = Field(default=None, alias="bidSource")
    delivery_date: Optional[datetime.date] = Field(None, alias="deliveryDate")
    bid_generated: Optional[datetime.datetime] = Field(None, alias="bidGenerated")

    def as_write(self) -> BenchmarkingShopCaseWrite:
        """Convert this read version of benchmarking shop case to the writing version."""
        return BenchmarkingShopCaseWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            scenario=self.scenario.as_write() if isinstance(self.scenario, DomainModel) else self.scenario,
            start_time=self.start_time,
            end_time=self.end_time,
            shop_files=[
                shop_file.as_write() if isinstance(shop_file, DomainModel) else shop_file
                for shop_file in self.shop_files or []
            ],
            bid_source=self.bid_source,
            delivery_date=self.delivery_date,
            bid_generated=self.bid_generated,
        )

    def as_apply(self) -> BenchmarkingShopCaseWrite:
        """Convert this read version of benchmarking shop case to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class BenchmarkingShopCaseWrite(ShopCaseWrite):
    """This represents the writing version of benchmarking shop case.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the benchmarking shop case.
        data_record: The data record of the benchmarking shop case node.
        scenario: The Shop scenario that was used to produce this result
        start_time: The start time of the case
        end_time: The end time of the case
        shop_files: The list of shop files that are used in a shop run. This encompasses all shop files such as case, module series, cut files etc.
        bid_source: The bid source field.
        delivery_date: The delivery date
        bid_generated: Timestamp of when the bid had been generated
    """

    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "power_ops_types", "BenchmarkingShopCase"
    )
    bid_source: Union[str, dm.NodeId, None] = Field(default=None, alias="bidSource")
    delivery_date: Optional[datetime.date] = Field(None, alias="deliveryDate")
    bid_generated: Optional[datetime.datetime] = Field(None, alias="bidGenerated")

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
        allow_version_increase: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(
            BenchmarkingShopCase, dm.ViewId("power_ops_core", "BenchmarkingShopCase", "1")
        )

        properties: dict[str, Any] = {}

        if self.scenario is not None:
            properties["scenario"] = {
                "space": self.space if isinstance(self.scenario, str) else self.scenario.space,
                "externalId": self.scenario if isinstance(self.scenario, str) else self.scenario.external_id,
            }

        if self.start_time is not None or write_none:
            properties["startTime"] = self.start_time.isoformat(timespec="milliseconds") if self.start_time else None

        if self.end_time is not None or write_none:
            properties["endTime"] = self.end_time.isoformat(timespec="milliseconds") if self.end_time else None

        if self.bid_source is not None:
            properties["bidSource"] = {
                "space": self.space if isinstance(self.bid_source, str) else self.bid_source.space,
                "externalId": self.bid_source if isinstance(self.bid_source, str) else self.bid_source.external_id,
            }

        if self.delivery_date is not None or write_none:
            properties["deliveryDate"] = self.delivery_date.isoformat() if self.delivery_date else None

        if self.bid_generated is not None or write_none:
            properties["bidGenerated"] = (
                self.bid_generated.isoformat(timespec="milliseconds") if self.bid_generated else None
            )

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=None if allow_version_increase else self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        edge_type = dm.DirectRelationReference("power_ops_types", "ShopCase.shopFiles")
        for shop_file in self.shop_files or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache,
                start_node=self,
                end_node=shop_file,
                edge_type=edge_type,
                view_by_read_class=view_by_read_class,
                write_none=write_none,
                allow_version_increase=allow_version_increase,
            )
            resources.extend(other_resources)

        if isinstance(self.scenario, DomainModelWrite):
            other_resources = self.scenario._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        return resources


class BenchmarkingShopCaseApply(BenchmarkingShopCaseWrite):
    def __new__(cls, *args, **kwargs) -> BenchmarkingShopCaseApply:
        warnings.warn(
            "BenchmarkingShopCaseApply is deprecated and will be removed in v1.0. Use BenchmarkingShopCaseWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "BenchmarkingShopCase.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class BenchmarkingShopCaseList(DomainModelList[BenchmarkingShopCase]):
    """List of benchmarking shop cases in the read version."""

    _INSTANCE = BenchmarkingShopCase

    def as_write(self) -> BenchmarkingShopCaseWriteList:
        """Convert these read versions of benchmarking shop case to the writing versions."""
        return BenchmarkingShopCaseWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> BenchmarkingShopCaseWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class BenchmarkingShopCaseWriteList(DomainModelWriteList[BenchmarkingShopCaseWrite]):
    """List of benchmarking shop cases in the writing version."""

    _INSTANCE = BenchmarkingShopCaseWrite


class BenchmarkingShopCaseApplyList(BenchmarkingShopCaseWriteList): ...


def _create_benchmarking_shop_case_filter(
    view_id: dm.ViewId,
    scenario: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    min_start_time: datetime.datetime | None = None,
    max_start_time: datetime.datetime | None = None,
    min_end_time: datetime.datetime | None = None,
    max_end_time: datetime.datetime | None = None,
    bid_source: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    min_delivery_date: datetime.date | None = None,
    max_delivery_date: datetime.date | None = None,
    min_bid_generated: datetime.datetime | None = None,
    max_bid_generated: datetime.datetime | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if scenario and isinstance(scenario, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("scenario"), value={"space": DEFAULT_INSTANCE_SPACE, "externalId": scenario}
            )
        )
    if scenario and isinstance(scenario, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("scenario"), value={"space": scenario[0], "externalId": scenario[1]}
            )
        )
    if scenario and isinstance(scenario, list) and isinstance(scenario[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("scenario"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in scenario],
            )
        )
    if scenario and isinstance(scenario, list) and isinstance(scenario[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("scenario"),
                values=[{"space": item[0], "externalId": item[1]} for item in scenario],
            )
        )
    if min_start_time is not None or max_start_time is not None:
        filters.append(
            dm.filters.Range(
                view_id.as_property_ref("startTime"),
                gte=min_start_time.isoformat(timespec="milliseconds") if min_start_time else None,
                lte=max_start_time.isoformat(timespec="milliseconds") if max_start_time else None,
            )
        )
    if min_end_time is not None or max_end_time is not None:
        filters.append(
            dm.filters.Range(
                view_id.as_property_ref("endTime"),
                gte=min_end_time.isoformat(timespec="milliseconds") if min_end_time else None,
                lte=max_end_time.isoformat(timespec="milliseconds") if max_end_time else None,
            )
        )
    if bid_source and isinstance(bid_source, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("bidSource"), value={"space": DEFAULT_INSTANCE_SPACE, "externalId": bid_source}
            )
        )
    if bid_source and isinstance(bid_source, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("bidSource"), value={"space": bid_source[0], "externalId": bid_source[1]}
            )
        )
    if bid_source and isinstance(bid_source, list) and isinstance(bid_source[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("bidSource"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in bid_source],
            )
        )
    if bid_source and isinstance(bid_source, list) and isinstance(bid_source[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("bidSource"),
                values=[{"space": item[0], "externalId": item[1]} for item in bid_source],
            )
        )
    if min_delivery_date is not None or max_delivery_date is not None:
        filters.append(
            dm.filters.Range(
                view_id.as_property_ref("deliveryDate"),
                gte=min_delivery_date.isoformat() if min_delivery_date else None,
                lte=max_delivery_date.isoformat() if max_delivery_date else None,
            )
        )
    if min_bid_generated is not None or max_bid_generated is not None:
        filters.append(
            dm.filters.Range(
                view_id.as_property_ref("bidGenerated"),
                gte=min_bid_generated.isoformat(timespec="milliseconds") if min_bid_generated else None,
                lte=max_bid_generated.isoformat(timespec="milliseconds") if max_bid_generated else None,
            )
        )
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

from collections.abc import Sequence
from typing import overload, Literal
import warnings

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.instances import InstanceAggregationResultList

from cognite.powerops.client._generated.v1.data_classes._core import DEFAULT_INSTANCE_SPACE
from cognite.powerops.client._generated.v1.data_classes import (
    DomainModelCore,
    DomainModelWrite,
    ResourcesWriteResult,
    ShopResult,
    ShopResultWrite,
    ShopResultFields,
    ShopResultList,
    ShopResultWriteList,
)
from cognite.powerops.client._generated.v1.data_classes._shop_result import (
    _SHOPRESULT_PROPERTIES_BY_FIELD,
    _create_shop_result_filter,
)
from ._core import (
    DEFAULT_LIMIT_READ,
    DEFAULT_QUERY_LIMIT,
    Aggregations,
    NodeAPI,
    SequenceNotStr,
    QueryStep,
    QueryBuilder,
)
from .shop_result_alerts import ShopResultAlertsAPI
from .shop_result_output_time_series import ShopResultOutputTimeSeriesAPI
from .shop_result_query import ShopResultQueryAPI


class ShopResultAPI(NodeAPI[ShopResult, ShopResultWrite, ShopResultList]):
    def __init__(self, client: CogniteClient, view_by_read_class: dict[type[DomainModelCore], dm.ViewId]):
        view_id = view_by_read_class[ShopResult]
        super().__init__(
            client=client,
            sources=view_id,
            class_type=ShopResult,
            class_list=ShopResultList,
            class_write_list=ShopResultWriteList,
            view_by_read_class=view_by_read_class,
        )
        self._view_id = view_id
        self.alerts_edge = ShopResultAlertsAPI(client)
        self.output_time_series_edge = ShopResultOutputTimeSeriesAPI(client)

    def __call__(
        self,
        case: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_QUERY_LIMIT,
        filter: dm.Filter | None = None,
    ) -> ShopResultQueryAPI[ShopResultList]:
        """Query starting at shop results.

        Args:
            case: The case to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of shop results to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            A query API for shop results.

        """
        has_data = dm.filters.HasData(views=[self._view_id])
        filter_ = _create_shop_result_filter(
            self._view_id,
            case,
            external_id_prefix,
            space,
            (filter and dm.filters.And(filter, has_data)) or has_data,
        )
        builder = QueryBuilder(ShopResultList)
        return ShopResultQueryAPI(self._client, builder, self._view_by_read_class, filter_, limit)

    def apply(
        self,
        shop_result: ShopResultWrite | Sequence[ShopResultWrite],
        replace: bool = False,
        write_none: bool = False,
    ) -> ResourcesWriteResult:
        """Add or update (upsert) shop results.

        Note: This method iterates through all nodes and timeseries linked to shop_result and creates them including the edges
        between the nodes. For example, if any of `alerts` or `output_time_series` are set, then these
        nodes as well as any nodes linked to them, and all the edges linking these nodes will be created.

        Args:
            shop_result: Shop result or sequence of shop results to upsert.
            replace (bool): How do we behave when a property value exists? Do we replace all matching and existing values with the supplied values (true)?
                Or should we merge in new values for properties together with the existing values (false)? Note: This setting applies for all nodes or edges specified in the ingestion call.
            write_none (bool): This method, will by default, skip properties that are set to None. However, if you want to set properties to None,
                you can set this parameter to True. Note this only applies to properties that are nullable.
        Returns:
            Created instance(s), i.e., nodes, edges, and time series.

        Examples:

            Create a new shop_result:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> from cognite.powerops.client._generated.v1.data_classes import ShopResultWrite
                >>> client = PowerOpsModelsV1Client()
                >>> shop_result = ShopResultWrite(external_id="my_shop_result", ...)
                >>> result = client.shop_result.apply(shop_result)

        """
        warnings.warn(
            "The .apply method is deprecated and will be removed in v1.0. "
            "Please use the .upsert method on the client instead. This means instead of "
            "`my_client.shop_result.apply(my_items)` please use `my_client.upsert(my_items)`."
            "The motivation is that all apply methods are the same, and having one apply method per API "
            " class encourages users to create items in small batches, which is inefficient."
            "In addition, .upsert method is more descriptive of what the method does.",
            UserWarning,
            stacklevel=2,
        )
        return self._apply(shop_result, replace, write_none)

    def delete(
        self, external_id: str | SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> dm.InstancesDeleteResult:
        """Delete one or more shop result.

        Args:
            external_id: External id of the shop result to delete.
            space: The space where all the shop result are located.

        Returns:
            The instance(s), i.e., nodes and edges which has been deleted. Empty list if nothing was deleted.

        Examples:

            Delete shop_result by id:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> client.shop_result.delete("my_shop_result")
        """
        warnings.warn(
            "The .delete method is deprecated and will be removed in v1.0. "
            "Please use the .delete method on the client instead. This means instead of "
            "`my_client.shop_result.delete(my_ids)` please use `my_client.delete(my_ids)`."
            "The motivation is that all delete methods are the same, and having one delete method per API "
            " class encourages users to delete items in small batches, which is inefficient.",
            UserWarning,
            stacklevel=2,
        )
        return self._delete(external_id, space)

    @overload
    def retrieve(self, external_id: str, space: str = DEFAULT_INSTANCE_SPACE) -> ShopResult | None: ...

    @overload
    def retrieve(self, external_id: SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE) -> ShopResultList: ...

    def retrieve(
        self, external_id: str | SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> ShopResult | ShopResultList | None:
        """Retrieve one or more shop results by id(s).

        Args:
            external_id: External id or list of external ids of the shop results.
            space: The space where all the shop results are located.

        Returns:
            The requested shop results.

        Examples:

            Retrieve shop_result by id:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> shop_result = client.shop_result.retrieve("my_shop_result")

        """
        return self._retrieve(
            external_id,
            space,
            retrieve_edges=True,
            edge_api_name_type_direction_view_id_penta=[
                (
                    self.alerts_edge,
                    "alerts",
                    dm.DirectRelationReference("power_ops_types", "calculationIssue"),
                    "outwards",
                    dm.ViewId("power_ops_core", "Alert", "1"),
                ),
                (
                    self.output_time_series_edge,
                    "output_time_series",
                    dm.DirectRelationReference("power_ops_types", "ShopResult.outputTimeSeries"),
                    "outwards",
                    dm.ViewId("power_ops_core", "ShopTimeSeries", "1"),
                ),
            ],
        )

    @overload
    def aggregate(
        self,
        aggregations: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: ShopResultFields | Sequence[ShopResultFields] | None = None,
        group_by: None = None,
        case: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue]: ...

    @overload
    def aggregate(
        self,
        aggregations: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: ShopResultFields | Sequence[ShopResultFields] | None = None,
        group_by: ShopResultFields | Sequence[ShopResultFields] = None,
        case: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> InstanceAggregationResultList: ...

    def aggregate(
        self,
        aggregate: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: ShopResultFields | Sequence[ShopResultFields] | None = None,
        group_by: ShopResultFields | Sequence[ShopResultFields] | None = None,
        case: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue] | InstanceAggregationResultList:
        """Aggregate data across shop results

        Args:
            aggregate: The aggregation to perform.
            property: The property to perform aggregation on.
            group_by: The property to group by when doing the aggregation.
            case: The case to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of shop results to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Aggregation results.

        Examples:

            Count shop results in space `my_space`:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> result = client.shop_result.aggregate("count", space="my_space")

        """

        filter_ = _create_shop_result_filter(
            self._view_id,
            case,
            external_id_prefix,
            space,
            filter,
        )
        return self._aggregate(
            self._view_id,
            aggregate,
            _SHOPRESULT_PROPERTIES_BY_FIELD,
            property,
            group_by,
            None,
            None,
            limit,
            filter_,
        )

    def histogram(
        self,
        property: ShopResultFields,
        interval: float,
        case: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> dm.aggregations.HistogramValue:
        """Produces histograms for shop results

        Args:
            property: The property to use as the value in the histogram.
            interval: The interval to use for the histogram bins.
            case: The case to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of shop results to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Bucketed histogram results.

        """
        filter_ = _create_shop_result_filter(
            self._view_id,
            case,
            external_id_prefix,
            space,
            filter,
        )
        return self._histogram(
            self._view_id,
            property,
            interval,
            _SHOPRESULT_PROPERTIES_BY_FIELD,
            None,
            None,
            limit,
            filter_,
        )

    def list(
        self,
        case: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
        sort_by: ShopResultFields | Sequence[ShopResultFields] | None = None,
        direction: Literal["ascending", "descending"] = "ascending",
        retrieve_edges: bool = True,
    ) -> ShopResultList:
        """List/filter shop results

        Args:
            case: The case to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of shop results to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.
            sort_by: The property to sort by.
            direction: The direction to sort by, either 'ascending' or 'descending'.
            retrieve_edges: Whether to retrieve `alerts` or `output_time_series` external ids for the shop results. Defaults to True.

        Returns:
            List of requested shop results

        Examples:

            List shop results and limit to 5:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> shop_results = client.shop_result.list(limit=5)

        """
        filter_ = _create_shop_result_filter(
            self._view_id,
            case,
            external_id_prefix,
            space,
            filter,
        )

        return self._list(
            limit=limit,
            filter=filter_,
            properties_by_field=_SHOPRESULT_PROPERTIES_BY_FIELD,
            sort_by=sort_by,
            direction=direction,
            retrieve_edges=retrieve_edges,
            edge_api_name_type_direction_view_id_penta=[
                (
                    self.alerts_edge,
                    "alerts",
                    dm.DirectRelationReference("power_ops_types", "calculationIssue"),
                    "outwards",
                    dm.ViewId("power_ops_core", "Alert", "1"),
                ),
                (
                    self.output_time_series_edge,
                    "output_time_series",
                    dm.DirectRelationReference("power_ops_types", "ShopResult.outputTimeSeries"),
                    "outwards",
                    dm.ViewId("power_ops_core", "ShopTimeSeries", "1"),
                ),
            ],
        )



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field
from pydantic import field_validator, model_validator

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    GraphQLCore,
    ResourcesWrite,
)

if TYPE_CHECKING:
    from ._benchmarking_production_obligation_day_ahead import (
        BenchmarkingProductionObligationDayAhead,
        BenchmarkingProductionObligationDayAheadGraphQL,
        BenchmarkingProductionObligationDayAheadWrite,
    )
    from ._power_asset import PowerAsset, PowerAssetGraphQL, PowerAssetWrite
    from ._shop_commands import ShopCommands, ShopCommandsGraphQL, ShopCommandsWrite
    from ._shop_model import ShopModel, ShopModelGraphQL, ShopModelWrite


__all__ = [
    "ShopModelWithAssets",
    "ShopModelWithAssetsWrite",
    "ShopModelWithAssetsApply",
    "ShopModelWithAssetsList",
    "ShopModelWithAssetsWriteList",
    "ShopModelWithAssetsApplyList",
]


class ShopModelWithAssetsGraphQL(GraphQLCore):
    """This represents the reading version of shop model with asset, used
    when data is retrieved from CDF using GraphQL.

    It is used when retrieving data from CDF using GraphQL.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop model with asset.
        data_record: The data record of the shop model with asset node.
        shop_model: The shop model that includes one water course for one shop run
        shop_commands: Commands for the shop file
        power_assets: A list of power assets covered by the Shop model. For a given bid document, we will select the partial bids for these assets, and calculate production obligation for these partial bids (summed up)
        production_obligations: It is possible to specify time series for production obligation - one benchmarking run will be set up for each of these time series. The intended use of this, is to specify the production obligation resulting from the submitted bid, or any other bid document not modelled within PowerOps
    """

    view_id = dm.ViewId("power_ops_core", "ShopModelWithAssets", "1")
    shop_model: Optional[ShopModelGraphQL] = Field(default=None, repr=False, alias="shopModel")
    shop_commands: Optional[ShopCommandsGraphQL] = Field(default=None, repr=False, alias="shopCommands")
    power_assets: Optional[list[PowerAssetGraphQL]] = Field(default=None, repr=False, alias="powerAssets")
    production_obligations: Optional[list[BenchmarkingProductionObligationDayAheadGraphQL]] = Field(
        default=None, repr=False, alias="productionObligations"
    )

    @model_validator(mode="before")
    def parse_data_record(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values
        if "lastUpdatedTime" in values or "createdTime" in values:
            values["dataRecord"] = DataRecordGraphQL(
                created_time=values.pop("createdTime", None),
                last_updated_time=values.pop("lastUpdatedTime", None),
            )
        return values

    @field_validator("shop_model", "shop_commands", "power_assets", "production_obligations", mode="before")
    def parse_graphql(cls, value: Any) -> Any:
        if not isinstance(value, dict):
            return value
        if "items" in value:
            return value["items"]
        return value

    def as_read(self) -> ShopModelWithAssets:
        """Convert this GraphQL format of shop model with asset to the reading format."""
        if self.data_record is None:
            raise ValueError("This object cannot be converted to a read format because it lacks a data record.")
        return ShopModelWithAssets(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecord(
                version=0,
                last_updated_time=self.data_record.last_updated_time,
                created_time=self.data_record.created_time,
            ),
            shop_model=self.shop_model.as_read() if isinstance(self.shop_model, GraphQLCore) else self.shop_model,
            shop_commands=(
                self.shop_commands.as_read() if isinstance(self.shop_commands, GraphQLCore) else self.shop_commands
            ),
            power_assets=[power_asset.as_read() for power_asset in self.power_assets or []],
            production_obligations=[
                production_obligation.as_read() for production_obligation in self.production_obligations or []
            ],
        )

    def as_write(self) -> ShopModelWithAssetsWrite:
        """Convert this GraphQL format of shop model with asset to the writing format."""
        return ShopModelWithAssetsWrite(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=0),
            shop_model=self.shop_model.as_write() if isinstance(self.shop_model, GraphQLCore) else self.shop_model,
            shop_commands=(
                self.shop_commands.as_write() if isinstance(self.shop_commands, GraphQLCore) else self.shop_commands
            ),
            power_assets=[power_asset.as_write() for power_asset in self.power_assets or []],
            production_obligations=[
                production_obligation.as_write() for production_obligation in self.production_obligations or []
            ],
        )


class ShopModelWithAssets(DomainModel):
    """This represents the reading version of shop model with asset.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop model with asset.
        data_record: The data record of the shop model with asset node.
        shop_model: The shop model that includes one water course for one shop run
        shop_commands: Commands for the shop file
        power_assets: A list of power assets covered by the Shop model. For a given bid document, we will select the partial bids for these assets, and calculate production obligation for these partial bids (summed up)
        production_obligations: It is possible to specify time series for production obligation - one benchmarking run will be set up for each of these time series. The intended use of this, is to specify the production obligation resulting from the submitted bid, or any other bid document not modelled within PowerOps
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "power_ops_types", "ShopModelWithAssets"
    )
    shop_model: Union[ShopModel, str, dm.NodeId, None] = Field(default=None, repr=False, alias="shopModel")
    shop_commands: Union[ShopCommands, str, dm.NodeId, None] = Field(default=None, repr=False, alias="shopCommands")
    power_assets: Union[list[PowerAsset], list[str], list[dm.NodeId], None] = Field(
        default=None, repr=False, alias="powerAssets"
    )
    production_obligations: Union[list[BenchmarkingProductionObligationDayAhead], list[str], list[dm.NodeId], None] = (
        Field(default=None, repr=False, alias="productionObligations")
    )

    def as_write(self) -> ShopModelWithAssetsWrite:
        """Convert this read version of shop model with asset to the writing version."""
        return ShopModelWithAssetsWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            shop_model=self.shop_model.as_write() if isinstance(self.shop_model, DomainModel) else self.shop_model,
            shop_commands=(
                self.shop_commands.as_write() if isinstance(self.shop_commands, DomainModel) else self.shop_commands
            ),
            power_assets=[
                power_asset.as_write() if isinstance(power_asset, DomainModel) else power_asset
                for power_asset in self.power_assets or []
            ],
            production_obligations=[
                (
                    production_obligation.as_write()
                    if isinstance(production_obligation, DomainModel)
                    else production_obligation
                )
                for production_obligation in self.production_obligations or []
            ],
        )

    def as_apply(self) -> ShopModelWithAssetsWrite:
        """Convert this read version of shop model with asset to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class ShopModelWithAssetsWrite(DomainModelWrite):
    """This represents the writing version of shop model with asset.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop model with asset.
        data_record: The data record of the shop model with asset node.
        shop_model: The shop model that includes one water course for one shop run
        shop_commands: Commands for the shop file
        power_assets: A list of power assets covered by the Shop model. For a given bid document, we will select the partial bids for these assets, and calculate production obligation for these partial bids (summed up)
        production_obligations: It is possible to specify time series for production obligation - one benchmarking run will be set up for each of these time series. The intended use of this, is to specify the production obligation resulting from the submitted bid, or any other bid document not modelled within PowerOps
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "power_ops_types", "ShopModelWithAssets"
    )
    shop_model: Union[ShopModelWrite, str, dm.NodeId, None] = Field(default=None, repr=False, alias="shopModel")
    shop_commands: Union[ShopCommandsWrite, str, dm.NodeId, None] = Field(
        default=None, repr=False, alias="shopCommands"
    )
    power_assets: Union[list[PowerAssetWrite], list[str], list[dm.NodeId], None] = Field(
        default=None, repr=False, alias="powerAssets"
    )
    production_obligations: Union[
        list[BenchmarkingProductionObligationDayAheadWrite], list[str], list[dm.NodeId], None
    ] = Field(default=None, repr=False, alias="productionObligations")

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
        allow_version_increase: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(
            ShopModelWithAssets, dm.ViewId("power_ops_core", "ShopModelWithAssets", "1")
        )

        properties: dict[str, Any] = {}

        if self.shop_model is not None:
            properties["shopModel"] = {
                "space": self.space if isinstance(self.shop_model, str) else self.shop_model.space,
                "externalId": self.shop_model if isinstance(self.shop_model, str) else self.shop_model.external_id,
            }

        if self.shop_commands is not None:
            properties["shopCommands"] = {
                "space": self.space if isinstance(self.shop_commands, str) else self.shop_commands.space,
                "externalId": (
                    self.shop_commands if isinstance(self.shop_commands, str) else self.shop_commands.external_id
                ),
            }

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=None if allow_version_increase else self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        edge_type = dm.DirectRelationReference("power_ops_core", "ShopModelWithAssets")
        for power_asset in self.power_assets or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache,
                start_node=self,
                end_node=power_asset,
                edge_type=edge_type,
                view_by_read_class=view_by_read_class,
                write_none=write_none,
                allow_version_increase=allow_version_increase,
            )
            resources.extend(other_resources)

        edge_type = dm.DirectRelationReference("power_ops_core", "ShopModelWithAssets")
        for production_obligation in self.production_obligations or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache,
                start_node=self,
                end_node=production_obligation,
                edge_type=edge_type,
                view_by_read_class=view_by_read_class,
                write_none=write_none,
                allow_version_increase=allow_version_increase,
            )
            resources.extend(other_resources)

        if isinstance(self.shop_model, DomainModelWrite):
            other_resources = self.shop_model._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        if isinstance(self.shop_commands, DomainModelWrite):
            other_resources = self.shop_commands._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        return resources


class ShopModelWithAssetsApply(ShopModelWithAssetsWrite):
    def __new__(cls, *args, **kwargs) -> ShopModelWithAssetsApply:
        warnings.warn(
            "ShopModelWithAssetsApply is deprecated and will be removed in v1.0. Use ShopModelWithAssetsWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "ShopModelWithAssets.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class ShopModelWithAssetsList(DomainModelList[ShopModelWithAssets]):
    """List of shop model with assets in the read version."""

    _INSTANCE = ShopModelWithAssets

    def as_write(self) -> ShopModelWithAssetsWriteList:
        """Convert these read versions of shop model with asset to the writing versions."""
        return ShopModelWithAssetsWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> ShopModelWithAssetsWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class ShopModelWithAssetsWriteList(DomainModelWriteList[ShopModelWithAssetsWrite]):
    """List of shop model with assets in the writing version."""

    _INSTANCE = ShopModelWithAssetsWrite


class ShopModelWithAssetsApplyList(ShopModelWithAssetsWriteList): ...


def _create_shop_model_with_asset_filter(
    view_id: dm.ViewId,
    shop_model: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    shop_commands: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if shop_model and isinstance(shop_model, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("shopModel"), value={"space": DEFAULT_INSTANCE_SPACE, "externalId": shop_model}
            )
        )
    if shop_model and isinstance(shop_model, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("shopModel"), value={"space": shop_model[0], "externalId": shop_model[1]}
            )
        )
    if shop_model and isinstance(shop_model, list) and isinstance(shop_model[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("shopModel"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in shop_model],
            )
        )
    if shop_model and isinstance(shop_model, list) and isinstance(shop_model[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("shopModel"),
                values=[{"space": item[0], "externalId": item[1]} for item in shop_model],
            )
        )
    if shop_commands and isinstance(shop_commands, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("shopCommands"),
                value={"space": DEFAULT_INSTANCE_SPACE, "externalId": shop_commands},
            )
        )
    if shop_commands and isinstance(shop_commands, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("shopCommands"),
                value={"space": shop_commands[0], "externalId": shop_commands[1]},
            )
        )
    if shop_commands and isinstance(shop_commands, list) and isinstance(shop_commands[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("shopCommands"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in shop_commands],
            )
        )
    if shop_commands and isinstance(shop_commands, list) and isinstance(shop_commands[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("shopCommands"),
                values=[{"space": item[0], "externalId": item[1]} for item in shop_commands],
            )
        )
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

import contextlib
import importlib
import json
import os
import re
import string
import warnings
from collections.abc import Iterator
from pathlib import Path
from typing import Annotated, Any, ForwardRef, Literal, Union, get_args, get_origin, overload

import tomli_w
from cognite.client.data_classes import TimeSeries
from cognite.client.utils._text import to_camel_case
from pydantic.v1.typing import evaluate_forwardref
from yaml import CSafeLoader, safe_dump

try:
    import tomllib  # type: ignore[import-not-found, unused-ignore]
except ModuleNotFoundError:  # py < 3.11
    import tomli as tomllib  # type: ignore[import-not-found, unused-ignore]


def read_toml_file(toml_file: Path | str) -> dict[str, Any]:
    """
    Read a toml file and return a dictionary.

    Args:
        toml_file: The path to the toml file.

    Returns:
        A dictionary with the toml data.
    """
    return tomllib.loads(Path(toml_file).read_text())


def dump_toml_file(toml_file: Path | str, data: dict[str, Any]) -> None:
    """
    Dump a dictionary to a toml file.

    Args:
        toml_file: The path to the toml file.
        data: The data to dump.

    """
    Path(toml_file).write_text(tomli_w.dumps(data))


@contextlib.contextmanager
def chdir(new_dir: Path) -> Iterator[None]:
    """
    Change directory to new_dir and return to the original directory when exiting the context.

    Args:
        new_dir: The new directory to change to.

    """
    current_working_dir = Path.cwd()
    os.chdir(new_dir)

    try:
        yield

    finally:
        os.chdir(current_working_dir)


@contextlib.contextmanager
def environment_variables(environ: dict[str, str]) -> Iterator[None]:
    """
    Temporarily set environment variables.

    Returns:
        None
    """
    original_environ = dict(os.environ)
    os.environ.update(environ)
    try:
        yield
    finally:
        os.environ.clear()
        os.environ.update(original_environ)


# � character is used to represent unrecognizable characters in utf-8.


# Having a single yaml dump function makes it easy to look up all places YaMLs are dumped.
# It also enables one place to set the default encoding.

UNRECOGNIZABLE_CHARACTER = "�"
VALID_CHARACTERS = set(
    string.ascii_lowercase
    + string.ascii_uppercase
    + string.digits
    + UNRECOGNIZABLE_CHARACTER
    + string.punctuation
    + string.whitespace
    + "æøåÆØÅ"
)
_READ_ONLY_FIELDS = [
    "created_time",
    "last_updated_time",
    "uploaded_time",
    "data_set_id",
    "id",
    "parent_id",
    "root_id",
    "uploaded",
]


def try_load_list(value: str | Any) -> Any | list[Any]:
    if isinstance(value, str):
        try:
            return json.loads(value)
        except json.JSONDecodeError:
            return []
    return value


def try_load_dict(value: str | Any) -> Any | dict[str, Any]:
    if isinstance(value, str):
        try:
            return json.loads(value)
        except json.JSONDecodeError:
            return {}
    return value


def parse_time_series(value: str | Any) -> Any | TimeSeries:
    if isinstance(value, TimeSeries) or value is None:
        return value
    elif value == {}:
        return None
    elif isinstance(value, dict):
        return TimeSeries._load(value)
    raise NotImplementedError()


def remove_read_only_fields(cdf_resource: dict[str, Any], remove_empty: bool = True) -> dict[str, Any]:
    for field in _READ_ONLY_FIELDS:
        cdf_resource.pop(field, None)
        cdf_resource.pop(to_camel_case(field), None)
    if remove_empty:
        for field in list(cdf_resource):
            if not cdf_resource[field]:
                cdf_resource.pop(field)
    return cdf_resource


def _validate(yaml_path: Path) -> None:
    if yaml_path.suffix not in {".yaml", ".yml"}:
        raise ValueError(f"File {yaml_path.name} not a valid yaml {yaml_path.suffix}")


@overload
def load_yaml(
    yaml_path: Path, expected_return_type: Literal["dict"] = "dict", encoding: str = "utf-8", clean_data: bool = False
) -> dict: ...


@overload
def load_yaml(
    yaml_path: Path, expected_return_type: Literal["list"], encoding: str = "utf-8", clean_data: bool = False
) -> list: ...


@overload
def load_yaml(
    yaml_path: Path, expected_return_type: Literal["any"], encoding: str = "utf-8", clean_data: bool = False
) -> list | dict: ...


def load_yaml(
    yaml_path: Path,
    expected_return_type: Literal["dict", "list", "any"] = "any",
    encoding: str = "utf-8",
    clean_data: bool = False,
) -> dict | list:
    """
    Fast loading of a yaml file.

    Args:
        yaml_path: The path to the yaml file.
        expected_return_type: The expected return type. The function will raise an error
                              if the file does not return the expected type. Defaults to any.
        encoding: The encoding of the yaml file. Defaults to utf-8.
        clean_data: Whether to clean the data from invalid characters. Defaults to False.

    Returns:
        The data in the yaml file as a dictionary.
    """

    _validate(yaml_path)
    # The Windows Cpython implementation seems to guess if encoding is not explicitly set
    # This turns out to be a problem as it guesses wrong, which is not the case for Unix systems.
    data = Path(yaml_path).read_text(encoding=encoding)

    if clean_data and (invalid_characters := (set(data) - VALID_CHARACTERS)):
        data = re.sub(rf"[{'|'.join(invalid_characters)}]", UNRECOGNIZABLE_CHARACTER, data)
        warnings.warn(
            f"File {yaml_path.parent}/{yaml_path.name} contains invalid characters: {', '.join(invalid_characters)}",
            stacklevel=2,
        )
    output = CSafeLoader(data).get_data()
    if expected_return_type == "dict" and not isinstance(output, dict):
        if not output:
            warnings.warn(
                f"File {yaml_path.parent}/{yaml_path.name} contains no data",
                stacklevel=2,
            )
            return {}
        raise ValueError(f"Expected a dictionary, got {type(output)}")
    if expected_return_type == "list" and not isinstance(output, list):
        if not output:
            warnings.warn(
                f"File {yaml_path.parent}/{yaml_path.name} contains no data",
                stacklevel=2,
            )
            return []
        raise ValueError(f"Expected a list, got {type(output)}")
    return output


def dump_yaml(yaml_path: Path, data: dict, encoding: str = "utf-8") -> None:
    """
    Dump a dictionary to a yaml file.

    Args:
        yaml_path: The path to the yaml file.
        data: The data to dump.
        encoding: The encoding of the yaml file. Defaults to utf-8.

    """
    _validate(yaml_path)
    with yaml_path.open("w", encoding=encoding) as stream:
        safe_dump(data, stream)


def get_pydantic_annotation(field_annotation: Any, cls_obj: type[object]) -> tuple[Any, type[dict] | type[list] | None]:
    if isinstance(field_annotation, ForwardRef):
        module = vars(importlib.import_module(cls_obj.__module__))
        parent_module = vars(importlib.import_module(cls_obj.__module__.rsplit(".", maxsplit=1)[0]))
        module.update(parent_module)
        field_annotation = evaluate_forwardref(field_annotation, globals(), module)

    outer: type[dict] | type[list] | None
    if not (origin := get_origin(field_annotation)):
        return field_annotation, None
    if origin is list:
        annotation, *_ = get_args(field_annotation)
        outer = list
    elif origin is dict:
        _, annotation = get_args(field_annotation)
        outer = dict
    elif origin is Union:
        annotation, *_ = get_args(field_annotation)
        return get_pydantic_annotation(annotation, cls_obj)
    elif origin is Annotated:
        return get_args(field_annotation)[0], None
    else:
        raise NotImplementedError(f"Cannot handle field_annotation  {field_annotation}")
    if inner := get_args(annotation):
        annotation = inner[0]
    return annotation, outer



import json
from typing import Any, Optional, Union

import arrow

from cognite.powerops.resync.config.market._core import RelativeTime


def relative_time_specification_to_arrow(
    relative_time_specification: Union[str, list[tuple[str, Any]]],
    floor_timezone: Optional[str] = None,
    result_timezone: str = "UTC",
) -> arrow.Arrow:
    """Get Arrow object based on utcnow and a relative time specification
    Required argument:
        relative_time_specification: List[Tuple[str, any]] or a JSON string representing the same - a list of tuples
                                     representing arrow methods with arguments. The functions will be called in the
                                     order they are specified in the list. The first element of the tuple is the arrow
                                     function name (can be "floor", "shift" and "to"), and the second element is the
                                     arguments to the function (str for "floor" and "to", kwargs Dict[str, int] for
                                     "shift").
    Optional arguments:
        floor_timezone: str representing a time zone - if specified, this .to(floor_timezone) will be called before the
                        functions given in relative_time_specification (to make e. g. .floor("day") floor to the day of
                        this timezone)
        result_timezone: str representing a time zone - if specified, the return arrow object will have this timezone
                         (.to(result_timezone) called after the functions in relative_time_specification), otherwise it
                         will be in UTC - this may be useful if the arrow object should be formatted to a string without
                         timezone specification afterwards
    Returns:
        arrow.Arrow date-time object (based on arrow.utcnow, modified with the specified functions, in result_timezone)
    Examples of relative time specification
        to get the start of tomorrow, Oslo time:
            [
                ("floor", "day"),
                ("shift", {"days": 1})
            ]
            this will be parsed as arrow.utcnow().to(<timezone from market config>).floor("day").shift(**{"days": 1})

        to get the end of Saturday next week (UTC):
        [
            ("to", "UTC"),
            ("floor", "week"),
            ("shift", {"weeks": 2}),
            ("shift", {"days": -1}),
        ]
        this will be parsed as
        arrow.utcnow().to(<tz from market config>).to("UTC").floor("week").shift(**{"weeks": 2}).shift(**{"days": -1})
    """
    arw = arrow.utcnow()  # TODO: better to pass this?
    if isinstance(relative_time_specification, str):
        relative_time_specification = json.loads(relative_time_specification)
        assert isinstance(relative_time_specification, list)
    if isinstance(relative_time_specification, RelativeTime):
        relative_time_specification = relative_time_specification.operations
        assert isinstance(relative_time_specification, list)
    if floor_timezone:  # TODO: should this be part of the spec instead?
        arw = arw.to(floor_timezone)
    for method, arg in relative_time_specification:
        if method == "to":
            arw = arw.to(arg)
        elif method == "floor":
            arw = arw.floor(arg)
        elif method == "shift":
            arw = arw.shift(**arg)
        else:
            print(f"Invalid relative time specification method: {method}")
    return arw.to(result_timezone)


def arrow_to_ms(value: arrow.Arrow) -> int:
    """Arrow datetime to milliseconds since Epoch."""
    return int(value.float_timestamp * 1000)



from __future__ import annotations


from cognite.client import data_modeling as dm

from ._core import DEFAULT_LIMIT_READ, EdgeAPI, _create_edge_filter
from cognite.powerops.client._generated.v1.data_classes._core import DEFAULT_INSTANCE_SPACE


class BenchmarkingShopCaseShopFilesAPI(EdgeAPI):
    def list(
        self,
        from_benchmarking_shop_case: str | list[str] | dm.NodeId | list[dm.NodeId] | None = None,
        from_benchmarking_shop_case_space: str = DEFAULT_INSTANCE_SPACE,
        to_shop_file: str | list[str] | dm.NodeId | list[dm.NodeId] | None = None,
        to_shop_file_space: str = DEFAULT_INSTANCE_SPACE,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit=DEFAULT_LIMIT_READ,
    ) -> dm.EdgeList:
        """List shop file edges of a benchmarking shop case.

        Args:
            from_benchmarking_shop_case: ID of the source benchmarking shop case.
            from_benchmarking_shop_case_space: Location of the benchmarking shop cases.
            to_shop_file: ID of the target shop file.
            to_shop_file_space: Location of the shop files.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of shop file edges to return. Defaults to 25. Set to -1, float("inf") or None
                to return all items.

        Returns:
            The requested shop file edges.

        Examples:

            List 5 shop file edges connected to "my_benchmarking_shop_case":

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> benchmarking_shop_case = client.benchmarking_shop_case.shop_files_edge.list("my_benchmarking_shop_case", limit=5)

        """
        filter_ = _create_edge_filter(
            dm.DirectRelationReference("power_ops_types", "ShopCase.shopFiles"),
            from_benchmarking_shop_case,
            from_benchmarking_shop_case_space,
            to_shop_file,
            to_shop_file_space,
            external_id_prefix,
            space,
        )
        return self._list(filter_=filter_, limit=limit)



from __future__ import annotations

from typing import Protocol


class Echo(Protocol):
    """
    Protocol for the echo function used by resync to communicate with the user.
    """

    def __call__(self, message: str, is_warning: bool = False) -> None:
        """
        Message printer.

        Args:
            message: The message to print.
            is_warning: Whether the message is a warning or not.

        """
        ...



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from cognite.client.data_classes import TimeSeries as CogniteTimeSeries
from pydantic import Field
from pydantic import field_validator, model_validator

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    GraphQLCore,
    ResourcesWrite,
    TimeSeries,
)

if TYPE_CHECKING:
    from ._generator_efficiency_curve import (
        GeneratorEfficiencyCurve,
        GeneratorEfficiencyCurveGraphQL,
        GeneratorEfficiencyCurveWrite,
    )
    from ._turbine_efficiency_curve import (
        TurbineEfficiencyCurve,
        TurbineEfficiencyCurveGraphQL,
        TurbineEfficiencyCurveWrite,
    )


__all__ = [
    "Generator",
    "GeneratorWrite",
    "GeneratorApply",
    "GeneratorList",
    "GeneratorWriteList",
    "GeneratorApplyList",
    "GeneratorFields",
    "GeneratorTextFields",
]


GeneratorTextFields = Literal["name", "display_name", "start_stop_cost", "is_available_time_series"]
GeneratorFields = Literal[
    "name", "display_name", "p_min", "penstock", "start_cost", "start_stop_cost", "is_available_time_series"
]

_GENERATOR_PROPERTIES_BY_FIELD = {
    "name": "name",
    "display_name": "displayName",
    "p_min": "pMin",
    "penstock": "penstock",
    "start_cost": "startCost",
    "start_stop_cost": "startStopCost",
    "is_available_time_series": "isAvailableTimeSeries",
}


class GeneratorGraphQL(GraphQLCore):
    """This represents the reading version of generator, used
    when data is retrieved from CDF using GraphQL.

    It is used when retrieving data from CDF using GraphQL.

    Args:
        space: The space where the node is located.
        external_id: The external id of the generator.
        data_record: The data record of the generator node.
        name: Name for the Generator.
        display_name: Display name for the Generator.
        p_min: The p min field.
        penstock: The penstock field.
        start_cost: The start cost field.
        start_stop_cost: The start stop cost field.
        is_available_time_series: The is available time series field.
        efficiency_curve: The efficiency curve field.
        turbine_curves: The watercourses that are connected to the PriceArea.
    """

    view_id = dm.ViewId("power-ops-assets", "Generator", "1")
    name: Optional[str] = None
    display_name: Optional[str] = Field(None, alias="displayName")
    p_min: Optional[float] = Field(None, alias="pMin")
    penstock: Optional[int] = None
    start_cost: Optional[float] = Field(None, alias="startCost")
    start_stop_cost: Union[TimeSeries, dict, None] = Field(None, alias="startStopCost")
    is_available_time_series: Union[TimeSeries, dict, None] = Field(None, alias="isAvailableTimeSeries")
    efficiency_curve: Optional[GeneratorEfficiencyCurveGraphQL] = Field(None, repr=False, alias="efficiencyCurve")
    turbine_curves: Optional[list[TurbineEfficiencyCurveGraphQL]] = Field(
        default=None, repr=False, alias="turbineCurves"
    )

    @model_validator(mode="before")
    def parse_data_record(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values
        if "lastUpdatedTime" in values or "createdTime" in values:
            values["dataRecord"] = DataRecordGraphQL(
                created_time=values.pop("createdTime", None),
                last_updated_time=values.pop("lastUpdatedTime", None),
            )
        return values

    @field_validator("efficiency_curve", "turbine_curves", mode="before")
    def parse_graphql(cls, value: Any) -> Any:
        if not isinstance(value, dict):
            return value
        if "items" in value:
            return value["items"]
        return value

    def as_read(self) -> Generator:
        """Convert this GraphQL format of generator to the reading format."""
        if self.data_record is None:
            raise ValueError("This object cannot be converted to a read format because it lacks a data record.")
        return Generator(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecord(
                version=0,
                last_updated_time=self.data_record.last_updated_time,
                created_time=self.data_record.created_time,
            ),
            name=self.name,
            display_name=self.display_name,
            p_min=self.p_min,
            penstock=self.penstock,
            start_cost=self.start_cost,
            start_stop_cost=self.start_stop_cost,
            is_available_time_series=self.is_available_time_series,
            efficiency_curve=(
                self.efficiency_curve.as_read()
                if isinstance(self.efficiency_curve, GraphQLCore)
                else self.efficiency_curve
            ),
            turbine_curves=[
                turbine_curve.as_read() if isinstance(turbine_curve, GraphQLCore) else turbine_curve
                for turbine_curve in self.turbine_curves or []
            ],
        )

    def as_write(self) -> GeneratorWrite:
        """Convert this GraphQL format of generator to the writing format."""
        return GeneratorWrite(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=0),
            name=self.name,
            display_name=self.display_name,
            p_min=self.p_min,
            penstock=self.penstock,
            start_cost=self.start_cost,
            start_stop_cost=self.start_stop_cost,
            is_available_time_series=self.is_available_time_series,
            efficiency_curve=(
                self.efficiency_curve.as_write()
                if isinstance(self.efficiency_curve, DomainModel)
                else self.efficiency_curve
            ),
            turbine_curves=[
                turbine_curve.as_write() if isinstance(turbine_curve, DomainModel) else turbine_curve
                for turbine_curve in self.turbine_curves or []
            ],
        )


class Generator(DomainModel):
    """This represents the reading version of generator.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the generator.
        data_record: The data record of the generator node.
        name: Name for the Generator.
        display_name: Display name for the Generator.
        p_min: The p min field.
        penstock: The penstock field.
        start_cost: The start cost field.
        start_stop_cost: The start stop cost field.
        is_available_time_series: The is available time series field.
        efficiency_curve: The efficiency curve field.
        turbine_curves: The watercourses that are connected to the PriceArea.
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("power-ops-types", "Generator")
    name: str
    display_name: Optional[str] = Field(None, alias="displayName")
    p_min: Optional[float] = Field(None, alias="pMin")
    penstock: Optional[int] = None
    start_cost: Optional[float] = Field(None, alias="startCost")
    start_stop_cost: Union[TimeSeries, str, None] = Field(None, alias="startStopCost")
    is_available_time_series: Union[TimeSeries, str, None] = Field(None, alias="isAvailableTimeSeries")
    efficiency_curve: Union[GeneratorEfficiencyCurve, str, dm.NodeId, None] = Field(
        None, repr=False, alias="efficiencyCurve"
    )
    turbine_curves: Union[list[TurbineEfficiencyCurve], list[str], list[dm.NodeId], None] = Field(
        default=None, repr=False, alias="turbineCurves"
    )

    def as_write(self) -> GeneratorWrite:
        """Convert this read version of generator to the writing version."""
        return GeneratorWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            name=self.name,
            display_name=self.display_name,
            p_min=self.p_min,
            penstock=self.penstock,
            start_cost=self.start_cost,
            start_stop_cost=self.start_stop_cost,
            is_available_time_series=self.is_available_time_series,
            efficiency_curve=(
                self.efficiency_curve.as_write()
                if isinstance(self.efficiency_curve, DomainModel)
                else self.efficiency_curve
            ),
            turbine_curves=[
                turbine_curve.as_write() if isinstance(turbine_curve, DomainModel) else turbine_curve
                for turbine_curve in self.turbine_curves or []
            ],
        )

    def as_apply(self) -> GeneratorWrite:
        """Convert this read version of generator to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class GeneratorWrite(DomainModelWrite):
    """This represents the writing version of generator.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the generator.
        data_record: The data record of the generator node.
        name: Name for the Generator.
        display_name: Display name for the Generator.
        p_min: The p min field.
        penstock: The penstock field.
        start_cost: The start cost field.
        start_stop_cost: The start stop cost field.
        is_available_time_series: The is available time series field.
        efficiency_curve: The efficiency curve field.
        turbine_curves: The watercourses that are connected to the PriceArea.
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("power-ops-types", "Generator")
    name: str
    display_name: Optional[str] = Field(None, alias="displayName")
    p_min: Optional[float] = Field(None, alias="pMin")
    penstock: Optional[int] = None
    start_cost: Optional[float] = Field(None, alias="startCost")
    start_stop_cost: Union[TimeSeries, str, None] = Field(None, alias="startStopCost")
    is_available_time_series: Union[TimeSeries, str, None] = Field(None, alias="isAvailableTimeSeries")
    efficiency_curve: Union[GeneratorEfficiencyCurveWrite, str, dm.NodeId, None] = Field(
        None, repr=False, alias="efficiencyCurve"
    )
    turbine_curves: Union[list[TurbineEfficiencyCurveWrite], list[str], list[dm.NodeId], None] = Field(
        default=None, repr=False, alias="turbineCurves"
    )

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
        allow_version_increase: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(Generator, dm.ViewId("power-ops-assets", "Generator", "1"))

        properties: dict[str, Any] = {}

        if self.name is not None:
            properties["name"] = self.name

        if self.display_name is not None or write_none:
            properties["displayName"] = self.display_name

        if self.p_min is not None or write_none:
            properties["pMin"] = self.p_min

        if self.penstock is not None or write_none:
            properties["penstock"] = self.penstock

        if self.start_cost is not None or write_none:
            properties["startCost"] = self.start_cost

        if self.start_stop_cost is not None or write_none:
            if isinstance(self.start_stop_cost, str) or self.start_stop_cost is None:
                properties["startStopCost"] = self.start_stop_cost
            else:
                properties["startStopCost"] = self.start_stop_cost.external_id

        if self.is_available_time_series is not None or write_none:
            if isinstance(self.is_available_time_series, str) or self.is_available_time_series is None:
                properties["isAvailableTimeSeries"] = self.is_available_time_series
            else:
                properties["isAvailableTimeSeries"] = self.is_available_time_series.external_id

        if self.efficiency_curve is not None:
            properties["efficiencyCurve"] = {
                "space": self.space if isinstance(self.efficiency_curve, str) else self.efficiency_curve.space,
                "externalId": (
                    self.efficiency_curve
                    if isinstance(self.efficiency_curve, str)
                    else self.efficiency_curve.external_id
                ),
            }

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=None if allow_version_increase else self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        edge_type = dm.DirectRelationReference("power-ops-types", "isSubAssetOf")
        for turbine_curve in self.turbine_curves or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache,
                start_node=self,
                end_node=turbine_curve,
                edge_type=edge_type,
                view_by_read_class=view_by_read_class,
                write_none=write_none,
                allow_version_increase=allow_version_increase,
            )
            resources.extend(other_resources)

        if isinstance(self.efficiency_curve, DomainModelWrite):
            other_resources = self.efficiency_curve._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        if isinstance(self.start_stop_cost, CogniteTimeSeries):
            resources.time_series.append(self.start_stop_cost)

        if isinstance(self.is_available_time_series, CogniteTimeSeries):
            resources.time_series.append(self.is_available_time_series)

        return resources


class GeneratorApply(GeneratorWrite):
    def __new__(cls, *args, **kwargs) -> GeneratorApply:
        warnings.warn(
            "GeneratorApply is deprecated and will be removed in v1.0. Use GeneratorWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "Generator.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class GeneratorList(DomainModelList[Generator]):
    """List of generators in the read version."""

    _INSTANCE = Generator

    def as_write(self) -> GeneratorWriteList:
        """Convert these read versions of generator to the writing versions."""
        return GeneratorWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> GeneratorWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class GeneratorWriteList(DomainModelWriteList[GeneratorWrite]):
    """List of generators in the writing version."""

    _INSTANCE = GeneratorWrite


class GeneratorApplyList(GeneratorWriteList): ...


def _create_generator_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    display_name: str | list[str] | None = None,
    display_name_prefix: str | None = None,
    min_p_min: float | None = None,
    max_p_min: float | None = None,
    min_penstock: int | None = None,
    max_penstock: int | None = None,
    min_start_cost: float | None = None,
    max_start_cost: float | None = None,
    efficiency_curve: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if isinstance(display_name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("displayName"), value=display_name))
    if display_name and isinstance(display_name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("displayName"), values=display_name))
    if display_name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("displayName"), value=display_name_prefix))
    if min_p_min is not None or max_p_min is not None:
        filters.append(dm.filters.Range(view_id.as_property_ref("pMin"), gte=min_p_min, lte=max_p_min))
    if min_penstock is not None or max_penstock is not None:
        filters.append(dm.filters.Range(view_id.as_property_ref("penstock"), gte=min_penstock, lte=max_penstock))
    if min_start_cost is not None or max_start_cost is not None:
        filters.append(dm.filters.Range(view_id.as_property_ref("startCost"), gte=min_start_cost, lte=max_start_cost))
    if efficiency_curve and isinstance(efficiency_curve, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("efficiencyCurve"),
                value={"space": DEFAULT_INSTANCE_SPACE, "externalId": efficiency_curve},
            )
        )
    if efficiency_curve and isinstance(efficiency_curve, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("efficiencyCurve"),
                value={"space": efficiency_curve[0], "externalId": efficiency_curve[1]},
            )
        )
    if efficiency_curve and isinstance(efficiency_curve, list) and isinstance(efficiency_curve[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("efficiencyCurve"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in efficiency_curve],
            )
        )
    if efficiency_curve and isinstance(efficiency_curve, list) and isinstance(efficiency_curve[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("efficiencyCurve"),
                values=[{"space": item[0], "externalId": item[1]} for item in efficiency_curve],
            )
        )
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

import datetime
from abc import abstractmethod
from collections import UserList
from collections.abc import Collection, Iterator, Mapping
from typing import Any, Callable, ClassVar, Generic, Optional, TypeVar, overload

import pandas as pd
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.instances import Properties, PropertyValue
from pydantic import BaseModel, Extra, Field, model_validator


class DomainModelCore(BaseModel):
    space: str
    external_id: str = Field(min_length=1, max_length=255)

    def id_tuple(self) -> tuple[str, str]:
        return self.space, self.external_id

    def to_pandas(self) -> pd.Series:
        return pd.Series(self.model_dump())

    def _repr_html_(self) -> str:
        """Returns HTML representation of DomainModel."""
        return self.to_pandas().to_frame("value")._repr_html_()  # type: ignore[operator]


T_TypeNodeCore = TypeVar("T_TypeNodeCore", bound=DomainModelCore)


class DomainModel(DomainModelCore):
    version: int
    last_updated_time: datetime.datetime
    created_time: datetime.datetime
    deleted_time: Optional[datetime.datetime] = None

    @classmethod
    def from_node(cls: type[T_TypeNode], node: dm.Node) -> T_TypeNode:
        data = node.dump(camel_case=False)
        # Extra unpacking to avoid crashing between core and property fields
        # can happen in there is a field named 'version' in the DominModel.
        return cls(**{**data, **unpack_properties(node.properties)})  # type: ignore[arg-type]


T_TypeNode = TypeVar("T_TypeNode", bound=DomainModel)


class DomainModelApply(DomainModelCore, extra=Extra.forbid, populate_by_name=True):
    external_id_factory: ClassVar[Optional[Callable[[type[DomainModelApply], dict], str]]] = None
    existing_version: Optional[int] = None

    def to_instances_apply(self) -> dm.InstancesApply:
        return self._to_instances_apply(set())

    @abstractmethod
    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        raise NotImplementedError()

    @model_validator(mode="before")
    def create_external_id_if_factory(cls, data: Any) -> Any:
        if isinstance(data, dict) and cls.external_id_factory is not None:
            data["external_id"] = cls.external_id_factory(cls, data)
        return data


T_TypeNodeApply = TypeVar("T_TypeNodeApply", bound=DomainModelApply)


class DomainModelApplyResult(DomainModelCore):
    version: int
    was_modified: bool
    last_updated_time: datetime.datetime
    created_time: datetime.datetime


class TypeList(UserList, Generic[T_TypeNodeCore]):
    _NODE: type[T_TypeNodeCore]

    def __init__(self, nodes: Collection[T_TypeNodeCore]):
        super().__init__(nodes)

    # The dunder implementations are to get proper type hints
    def __iter__(self) -> Iterator[T_TypeNodeCore]:
        return super().__iter__()

    @overload
    def __getitem__(self, item: int) -> T_TypeNodeCore: ...

    @overload
    def __getitem__(self: type[T_TypeNodeList], item: slice) -> T_TypeNodeList: ...

    def __getitem__(self, item: int | slice) -> T_TypeNodeCore | T_TypeNodeList:
        if isinstance(item, slice):
            return self.__class__(self.data[item])
        elif isinstance(item, int):
            return self.data[item]
        else:
            raise TypeError(f"Expected int or slice, got {type(item)}")

    def dump(self) -> list[dict[str, Any]]:
        return [node.model_dump() for node in self.data]

    def as_external_ids(self) -> list[str]:
        return [node.external_id for node in self.data]

    def as_node_ids(self) -> list[dm.NodeId]:
        return [dm.NodeId(space=node.space, external_id=node.external_id) for node in self.data]

    def to_pandas(self) -> pd.DataFrame:
        df = pd.DataFrame(self.dump())
        if df.empty:
            df = pd.DataFrame(columns=self._NODE.model_fields)
        # Reorder columns to have the custom columns first
        fixed_columns = set(DomainModel.model_fields)
        columns = (
            ["external_id"]
            + [col for col in df if col not in fixed_columns]
            + [col for col in DomainModel.model_fields if col != "external_id"]
        )
        return df[columns]

    def _repr_html_(self) -> str:
        return self.to_pandas()._repr_html_()  # type: ignore[operator]


T_TypeApplyNode = TypeVar("T_TypeApplyNode", bound=DomainModelApply)
T_TypeNodeList = TypeVar("T_TypeNodeList", bound=TypeList, covariant=True)


class TypeApplyList(TypeList[T_TypeApplyNode]):
    def to_instances_apply(self) -> dm.InstancesApply:
        cache: set[str] = set()
        nodes: list[dm.NodeApply] = []
        edges: list[dm.EdgeApply] = []
        for node in self.data:
            result = node._to_instances_apply(cache)
            nodes.extend(result.nodes)
            edges.extend(result.edges)
        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))


def unpack_properties(properties: Properties) -> Mapping[str, PropertyValue]:
    unpacked: dict[str, PropertyValue] = {}
    for view_properties in properties.values():
        for prop_name, prop_value in view_properties.items():
            if isinstance(prop_value, (str, int, float, bool, list)):
                unpacked[prop_name] = prop_value
            elif isinstance(prop_value, dict) and "externalId" in prop_value and "space" in prop_value:
                # Assumed to be reference properties
                unpacked[prop_name] = prop_value["externalId"]
            else:
                # JSON field
                unpacked[prop_name] = prop_value
    return unpacked



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field
from pydantic import field_validator, model_validator

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    GraphQLCore,
    ResourcesWrite,
)

if TYPE_CHECKING:
    from ._date_specification import DateSpecification, DateSpecificationGraphQL, DateSpecificationWrite
    from ._shop_scenario import ShopScenario, ShopScenarioGraphQL, ShopScenarioWrite


__all__ = [
    "ShopScenarioSet",
    "ShopScenarioSetWrite",
    "ShopScenarioSetApply",
    "ShopScenarioSetList",
    "ShopScenarioSetWriteList",
    "ShopScenarioSetApplyList",
    "ShopScenarioSetFields",
    "ShopScenarioSetTextFields",
]


ShopScenarioSetTextFields = Literal["name"]
ShopScenarioSetFields = Literal["name"]

_SHOPSCENARIOSET_PROPERTIES_BY_FIELD = {
    "name": "name",
}


class ShopScenarioSetGraphQL(GraphQLCore):
    """This represents the reading version of shop scenario set, used
    when data is retrieved from CDF using GraphQL.

    It is used when retrieving data from CDF using GraphQL.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop scenario set.
        data_record: The data record of the shop scenario set node.
        name: The name of the scenario set to run
        start_specification: TODO description
        end_specification: TODO description
        scenarios: Configuration of the partial bids that make up the total bid configuration
    """

    view_id = dm.ViewId("power_ops_core", "ShopScenarioSet", "1")
    name: Optional[str] = None
    start_specification: Optional[DateSpecificationGraphQL] = Field(
        default=None, repr=False, alias="startSpecification"
    )
    end_specification: Optional[DateSpecificationGraphQL] = Field(default=None, repr=False, alias="endSpecification")
    scenarios: Optional[list[ShopScenarioGraphQL]] = Field(default=None, repr=False)

    @model_validator(mode="before")
    def parse_data_record(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values
        if "lastUpdatedTime" in values or "createdTime" in values:
            values["dataRecord"] = DataRecordGraphQL(
                created_time=values.pop("createdTime", None),
                last_updated_time=values.pop("lastUpdatedTime", None),
            )
        return values

    @field_validator("start_specification", "end_specification", "scenarios", mode="before")
    def parse_graphql(cls, value: Any) -> Any:
        if not isinstance(value, dict):
            return value
        if "items" in value:
            return value["items"]
        return value

    def as_read(self) -> ShopScenarioSet:
        """Convert this GraphQL format of shop scenario set to the reading format."""
        if self.data_record is None:
            raise ValueError("This object cannot be converted to a read format because it lacks a data record.")
        return ShopScenarioSet(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecord(
                version=0,
                last_updated_time=self.data_record.last_updated_time,
                created_time=self.data_record.created_time,
            ),
            name=self.name,
            start_specification=(
                self.start_specification.as_read()
                if isinstance(self.start_specification, GraphQLCore)
                else self.start_specification
            ),
            end_specification=(
                self.end_specification.as_read()
                if isinstance(self.end_specification, GraphQLCore)
                else self.end_specification
            ),
            scenarios=[scenario.as_read() for scenario in self.scenarios or []],
        )

    def as_write(self) -> ShopScenarioSetWrite:
        """Convert this GraphQL format of shop scenario set to the writing format."""
        return ShopScenarioSetWrite(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=0),
            name=self.name,
            start_specification=(
                self.start_specification.as_write()
                if isinstance(self.start_specification, GraphQLCore)
                else self.start_specification
            ),
            end_specification=(
                self.end_specification.as_write()
                if isinstance(self.end_specification, GraphQLCore)
                else self.end_specification
            ),
            scenarios=[scenario.as_write() for scenario in self.scenarios or []],
        )


class ShopScenarioSet(DomainModel):
    """This represents the reading version of shop scenario set.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop scenario set.
        data_record: The data record of the shop scenario set node.
        name: The name of the scenario set to run
        start_specification: TODO description
        end_specification: TODO description
        scenarios: Configuration of the partial bids that make up the total bid configuration
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = None
    name: str
    start_specification: Union[DateSpecification, str, dm.NodeId, None] = Field(
        default=None, repr=False, alias="startSpecification"
    )
    end_specification: Union[DateSpecification, str, dm.NodeId, None] = Field(
        default=None, repr=False, alias="endSpecification"
    )
    scenarios: Union[list[ShopScenario], list[str], list[dm.NodeId], None] = Field(default=None, repr=False)

    def as_write(self) -> ShopScenarioSetWrite:
        """Convert this read version of shop scenario set to the writing version."""
        return ShopScenarioSetWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            name=self.name,
            start_specification=(
                self.start_specification.as_write()
                if isinstance(self.start_specification, DomainModel)
                else self.start_specification
            ),
            end_specification=(
                self.end_specification.as_write()
                if isinstance(self.end_specification, DomainModel)
                else self.end_specification
            ),
            scenarios=[
                scenario.as_write() if isinstance(scenario, DomainModel) else scenario
                for scenario in self.scenarios or []
            ],
        )

    def as_apply(self) -> ShopScenarioSetWrite:
        """Convert this read version of shop scenario set to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class ShopScenarioSetWrite(DomainModelWrite):
    """This represents the writing version of shop scenario set.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop scenario set.
        data_record: The data record of the shop scenario set node.
        name: The name of the scenario set to run
        start_specification: TODO description
        end_specification: TODO description
        scenarios: Configuration of the partial bids that make up the total bid configuration
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = None
    name: str
    start_specification: Union[DateSpecificationWrite, str, dm.NodeId, None] = Field(
        default=None, repr=False, alias="startSpecification"
    )
    end_specification: Union[DateSpecificationWrite, str, dm.NodeId, None] = Field(
        default=None, repr=False, alias="endSpecification"
    )
    scenarios: Union[list[ShopScenarioWrite], list[str], list[dm.NodeId], None] = Field(default=None, repr=False)

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
        allow_version_increase: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(
            ShopScenarioSet, dm.ViewId("power_ops_core", "ShopScenarioSet", "1")
        )

        properties: dict[str, Any] = {}

        if self.name is not None:
            properties["name"] = self.name

        if self.start_specification is not None:
            properties["startSpecification"] = {
                "space": self.space if isinstance(self.start_specification, str) else self.start_specification.space,
                "externalId": (
                    self.start_specification
                    if isinstance(self.start_specification, str)
                    else self.start_specification.external_id
                ),
            }

        if self.end_specification is not None:
            properties["endSpecification"] = {
                "space": self.space if isinstance(self.end_specification, str) else self.end_specification.space,
                "externalId": (
                    self.end_specification
                    if isinstance(self.end_specification, str)
                    else self.end_specification.external_id
                ),
            }

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=None if allow_version_increase else self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        edge_type = dm.DirectRelationReference("power_ops_types", "ShopScenarioSet.scenarios")
        for scenario in self.scenarios or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache,
                start_node=self,
                end_node=scenario,
                edge_type=edge_type,
                view_by_read_class=view_by_read_class,
                write_none=write_none,
                allow_version_increase=allow_version_increase,
            )
            resources.extend(other_resources)

        if isinstance(self.start_specification, DomainModelWrite):
            other_resources = self.start_specification._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        if isinstance(self.end_specification, DomainModelWrite):
            other_resources = self.end_specification._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        return resources


class ShopScenarioSetApply(ShopScenarioSetWrite):
    def __new__(cls, *args, **kwargs) -> ShopScenarioSetApply:
        warnings.warn(
            "ShopScenarioSetApply is deprecated and will be removed in v1.0. Use ShopScenarioSetWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "ShopScenarioSet.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class ShopScenarioSetList(DomainModelList[ShopScenarioSet]):
    """List of shop scenario sets in the read version."""

    _INSTANCE = ShopScenarioSet

    def as_write(self) -> ShopScenarioSetWriteList:
        """Convert these read versions of shop scenario set to the writing versions."""
        return ShopScenarioSetWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> ShopScenarioSetWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class ShopScenarioSetWriteList(DomainModelWriteList[ShopScenarioSetWrite]):
    """List of shop scenario sets in the writing version."""

    _INSTANCE = ShopScenarioSetWrite


class ShopScenarioSetApplyList(ShopScenarioSetWriteList): ...


def _create_shop_scenario_set_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    start_specification: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    end_specification: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if start_specification and isinstance(start_specification, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("startSpecification"),
                value={"space": DEFAULT_INSTANCE_SPACE, "externalId": start_specification},
            )
        )
    if start_specification and isinstance(start_specification, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("startSpecification"),
                value={"space": start_specification[0], "externalId": start_specification[1]},
            )
        )
    if start_specification and isinstance(start_specification, list) and isinstance(start_specification[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("startSpecification"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in start_specification],
            )
        )
    if start_specification and isinstance(start_specification, list) and isinstance(start_specification[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("startSpecification"),
                values=[{"space": item[0], "externalId": item[1]} for item in start_specification],
            )
        )
    if end_specification and isinstance(end_specification, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("endSpecification"),
                value={"space": DEFAULT_INSTANCE_SPACE, "externalId": end_specification},
            )
        )
    if end_specification and isinstance(end_specification, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("endSpecification"),
                value={"space": end_specification[0], "externalId": end_specification[1]},
            )
        )
    if end_specification and isinstance(end_specification, list) and isinstance(end_specification[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("endSpecification"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in end_specification],
            )
        )
    if end_specification and isinstance(end_specification, list) and isinstance(end_specification[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("endSpecification"),
                values=[{"space": item[0], "externalId": item[1]} for item in end_specification],
            )
        )
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field
from pydantic import field_validator, model_validator

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    GraphQLCore,
    ResourcesWrite,
)
from ._function_input import FunctionInput, FunctionInputWrite

if TYPE_CHECKING:
    from ._shop_case import ShopCase, ShopCaseGraphQL, ShopCaseWrite
    from ._shop_preprocessor_input import (
        ShopPreprocessorInput,
        ShopPreprocessorInputGraphQL,
        ShopPreprocessorInputWrite,
    )


__all__ = [
    "ShopTriggerInput",
    "ShopTriggerInputWrite",
    "ShopTriggerInputApply",
    "ShopTriggerInputList",
    "ShopTriggerInputWriteList",
    "ShopTriggerInputApplyList",
    "ShopTriggerInputFields",
    "ShopTriggerInputTextFields",
]


ShopTriggerInputTextFields = Literal["workflow_execution_id", "function_name", "function_call_id", "cog_shop_tag"]
ShopTriggerInputFields = Literal[
    "workflow_execution_id", "workflow_step", "function_name", "function_call_id", "cog_shop_tag"
]

_SHOPTRIGGERINPUT_PROPERTIES_BY_FIELD = {
    "workflow_execution_id": "workflowExecutionId",
    "workflow_step": "workflowStep",
    "function_name": "functionName",
    "function_call_id": "functionCallId",
    "cog_shop_tag": "cogShopTag",
}


class ShopTriggerInputGraphQL(GraphQLCore):
    """This represents the reading version of shop trigger input, used
    when data is retrieved from CDF using GraphQL.

    It is used when retrieving data from CDF using GraphQL.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop trigger input.
        data_record: The data record of the shop trigger input node.
        workflow_execution_id: The process associated with the function execution
        workflow_step: This is the step in the process.
        function_name: The name of the function
        function_call_id: The function call id
        cog_shop_tag: Optionally specify cogshop tag to trigger
        preprocessor_input: The preprocessor input to the shop run
        case: The SHOP case (with all details like model, scenario, and time series)
    """

    view_id = dm.ViewId("power_ops_core", "ShopTriggerInput", "1")
    workflow_execution_id: Optional[str] = Field(None, alias="workflowExecutionId")
    workflow_step: Optional[int] = Field(None, alias="workflowStep")
    function_name: Optional[str] = Field(None, alias="functionName")
    function_call_id: Optional[str] = Field(None, alias="functionCallId")
    cog_shop_tag: Optional[str] = Field(None, alias="cogShopTag")
    preprocessor_input: Optional[ShopPreprocessorInputGraphQL] = Field(
        default=None, repr=False, alias="preprocessorInput"
    )
    case: Optional[ShopCaseGraphQL] = Field(default=None, repr=False)

    @model_validator(mode="before")
    def parse_data_record(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values
        if "lastUpdatedTime" in values or "createdTime" in values:
            values["dataRecord"] = DataRecordGraphQL(
                created_time=values.pop("createdTime", None),
                last_updated_time=values.pop("lastUpdatedTime", None),
            )
        return values

    @field_validator("preprocessor_input", "case", mode="before")
    def parse_graphql(cls, value: Any) -> Any:
        if not isinstance(value, dict):
            return value
        if "items" in value:
            return value["items"]
        return value

    def as_read(self) -> ShopTriggerInput:
        """Convert this GraphQL format of shop trigger input to the reading format."""
        if self.data_record is None:
            raise ValueError("This object cannot be converted to a read format because it lacks a data record.")
        return ShopTriggerInput(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecord(
                version=0,
                last_updated_time=self.data_record.last_updated_time,
                created_time=self.data_record.created_time,
            ),
            workflow_execution_id=self.workflow_execution_id,
            workflow_step=self.workflow_step,
            function_name=self.function_name,
            function_call_id=self.function_call_id,
            cog_shop_tag=self.cog_shop_tag,
            preprocessor_input=(
                self.preprocessor_input.as_read()
                if isinstance(self.preprocessor_input, GraphQLCore)
                else self.preprocessor_input
            ),
            case=self.case.as_read() if isinstance(self.case, GraphQLCore) else self.case,
        )

    def as_write(self) -> ShopTriggerInputWrite:
        """Convert this GraphQL format of shop trigger input to the writing format."""
        return ShopTriggerInputWrite(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=0),
            workflow_execution_id=self.workflow_execution_id,
            workflow_step=self.workflow_step,
            function_name=self.function_name,
            function_call_id=self.function_call_id,
            cog_shop_tag=self.cog_shop_tag,
            preprocessor_input=(
                self.preprocessor_input.as_write()
                if isinstance(self.preprocessor_input, GraphQLCore)
                else self.preprocessor_input
            ),
            case=self.case.as_write() if isinstance(self.case, GraphQLCore) else self.case,
        )


class ShopTriggerInput(FunctionInput):
    """This represents the reading version of shop trigger input.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop trigger input.
        data_record: The data record of the shop trigger input node.
        workflow_execution_id: The process associated with the function execution
        workflow_step: This is the step in the process.
        function_name: The name of the function
        function_call_id: The function call id
        cog_shop_tag: Optionally specify cogshop tag to trigger
        preprocessor_input: The preprocessor input to the shop run
        case: The SHOP case (with all details like model, scenario, and time series)
    """

    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "power_ops_types", "ShopTriggerInput"
    )
    cog_shop_tag: Optional[str] = Field(None, alias="cogShopTag")
    preprocessor_input: Union[ShopPreprocessorInput, str, dm.NodeId, None] = Field(
        default=None, repr=False, alias="preprocessorInput"
    )
    case: Union[ShopCase, str, dm.NodeId, None] = Field(default=None, repr=False)

    def as_write(self) -> ShopTriggerInputWrite:
        """Convert this read version of shop trigger input to the writing version."""
        return ShopTriggerInputWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            workflow_execution_id=self.workflow_execution_id,
            workflow_step=self.workflow_step,
            function_name=self.function_name,
            function_call_id=self.function_call_id,
            cog_shop_tag=self.cog_shop_tag,
            preprocessor_input=(
                self.preprocessor_input.as_write()
                if isinstance(self.preprocessor_input, DomainModel)
                else self.preprocessor_input
            ),
            case=self.case.as_write() if isinstance(self.case, DomainModel) else self.case,
        )

    def as_apply(self) -> ShopTriggerInputWrite:
        """Convert this read version of shop trigger input to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class ShopTriggerInputWrite(FunctionInputWrite):
    """This represents the writing version of shop trigger input.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop trigger input.
        data_record: The data record of the shop trigger input node.
        workflow_execution_id: The process associated with the function execution
        workflow_step: This is the step in the process.
        function_name: The name of the function
        function_call_id: The function call id
        cog_shop_tag: Optionally specify cogshop tag to trigger
        preprocessor_input: The preprocessor input to the shop run
        case: The SHOP case (with all details like model, scenario, and time series)
    """

    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "power_ops_types", "ShopTriggerInput"
    )
    cog_shop_tag: Optional[str] = Field(None, alias="cogShopTag")
    preprocessor_input: Union[ShopPreprocessorInputWrite, str, dm.NodeId, None] = Field(
        default=None, repr=False, alias="preprocessorInput"
    )
    case: Union[ShopCaseWrite, str, dm.NodeId, None] = Field(default=None, repr=False)

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
        allow_version_increase: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(
            ShopTriggerInput, dm.ViewId("power_ops_core", "ShopTriggerInput", "1")
        )

        properties: dict[str, Any] = {}

        if self.workflow_execution_id is not None:
            properties["workflowExecutionId"] = self.workflow_execution_id

        if self.workflow_step is not None:
            properties["workflowStep"] = self.workflow_step

        if self.function_name is not None:
            properties["functionName"] = self.function_name

        if self.function_call_id is not None:
            properties["functionCallId"] = self.function_call_id

        if self.cog_shop_tag is not None or write_none:
            properties["cogShopTag"] = self.cog_shop_tag

        if self.preprocessor_input is not None:
            properties["preprocessorInput"] = {
                "space": self.space if isinstance(self.preprocessor_input, str) else self.preprocessor_input.space,
                "externalId": (
                    self.preprocessor_input
                    if isinstance(self.preprocessor_input, str)
                    else self.preprocessor_input.external_id
                ),
            }

        if self.case is not None:
            properties["case"] = {
                "space": self.space if isinstance(self.case, str) else self.case.space,
                "externalId": self.case if isinstance(self.case, str) else self.case.external_id,
            }

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=None if allow_version_increase else self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        if isinstance(self.preprocessor_input, DomainModelWrite):
            other_resources = self.preprocessor_input._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        if isinstance(self.case, DomainModelWrite):
            other_resources = self.case._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        return resources


class ShopTriggerInputApply(ShopTriggerInputWrite):
    def __new__(cls, *args, **kwargs) -> ShopTriggerInputApply:
        warnings.warn(
            "ShopTriggerInputApply is deprecated and will be removed in v1.0. Use ShopTriggerInputWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "ShopTriggerInput.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class ShopTriggerInputList(DomainModelList[ShopTriggerInput]):
    """List of shop trigger inputs in the read version."""

    _INSTANCE = ShopTriggerInput

    def as_write(self) -> ShopTriggerInputWriteList:
        """Convert these read versions of shop trigger input to the writing versions."""
        return ShopTriggerInputWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> ShopTriggerInputWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class ShopTriggerInputWriteList(DomainModelWriteList[ShopTriggerInputWrite]):
    """List of shop trigger inputs in the writing version."""

    _INSTANCE = ShopTriggerInputWrite


class ShopTriggerInputApplyList(ShopTriggerInputWriteList): ...


def _create_shop_trigger_input_filter(
    view_id: dm.ViewId,
    workflow_execution_id: str | list[str] | None = None,
    workflow_execution_id_prefix: str | None = None,
    min_workflow_step: int | None = None,
    max_workflow_step: int | None = None,
    function_name: str | list[str] | None = None,
    function_name_prefix: str | None = None,
    function_call_id: str | list[str] | None = None,
    function_call_id_prefix: str | None = None,
    cog_shop_tag: str | list[str] | None = None,
    cog_shop_tag_prefix: str | None = None,
    preprocessor_input: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    case: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if isinstance(workflow_execution_id, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("workflowExecutionId"), value=workflow_execution_id))
    if workflow_execution_id and isinstance(workflow_execution_id, list):
        filters.append(dm.filters.In(view_id.as_property_ref("workflowExecutionId"), values=workflow_execution_id))
    if workflow_execution_id_prefix is not None:
        filters.append(
            dm.filters.Prefix(view_id.as_property_ref("workflowExecutionId"), value=workflow_execution_id_prefix)
        )
    if min_workflow_step is not None or max_workflow_step is not None:
        filters.append(
            dm.filters.Range(view_id.as_property_ref("workflowStep"), gte=min_workflow_step, lte=max_workflow_step)
        )
    if isinstance(function_name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("functionName"), value=function_name))
    if function_name and isinstance(function_name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("functionName"), values=function_name))
    if function_name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("functionName"), value=function_name_prefix))
    if isinstance(function_call_id, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("functionCallId"), value=function_call_id))
    if function_call_id and isinstance(function_call_id, list):
        filters.append(dm.filters.In(view_id.as_property_ref("functionCallId"), values=function_call_id))
    if function_call_id_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("functionCallId"), value=function_call_id_prefix))
    if isinstance(cog_shop_tag, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("cogShopTag"), value=cog_shop_tag))
    if cog_shop_tag and isinstance(cog_shop_tag, list):
        filters.append(dm.filters.In(view_id.as_property_ref("cogShopTag"), values=cog_shop_tag))
    if cog_shop_tag_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("cogShopTag"), value=cog_shop_tag_prefix))
    if preprocessor_input and isinstance(preprocessor_input, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("preprocessorInput"),
                value={"space": DEFAULT_INSTANCE_SPACE, "externalId": preprocessor_input},
            )
        )
    if preprocessor_input and isinstance(preprocessor_input, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("preprocessorInput"),
                value={"space": preprocessor_input[0], "externalId": preprocessor_input[1]},
            )
        )
    if preprocessor_input and isinstance(preprocessor_input, list) and isinstance(preprocessor_input[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("preprocessorInput"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in preprocessor_input],
            )
        )
    if preprocessor_input and isinstance(preprocessor_input, list) and isinstance(preprocessor_input[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("preprocessorInput"),
                values=[{"space": item[0], "externalId": item[1]} for item in preprocessor_input],
            )
        )
    if case and isinstance(case, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("case"), value={"space": DEFAULT_INSTANCE_SPACE, "externalId": case}
            )
        )
    if case and isinstance(case, tuple):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("case"), value={"space": case[0], "externalId": case[1]})
        )
    if case and isinstance(case, list) and isinstance(case[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("case"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in case],
            )
        )
    if case and isinstance(case, list) and isinstance(case[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("case"), values=[{"space": item[0], "externalId": item[1]} for item in case]
            )
        )
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

from cognite.powerops.resync import models
from cognite.powerops.resync.config._main import ReSyncConfig
from cognite.powerops.resync.models.base import AssetModel, DataModel, Model
from cognite.powerops.resync.models.v1.config_to_model import (
    to_cogshop_asset_model,
    to_market_asset_model,
    to_production_model,
)
from cognite.powerops.resync.models.v2.config_to_model import to_powerasset_model


def transform(config: ReSyncConfig, market_name: str, model_types: set[type[Model]]) -> list[Model]:
    all_models: list[Model] = []

    # The Production model is a prerequisite for the Market and CogShop models
    has_asset_model = any(issubclass(m, (AssetModel, models.CogShop1Asset)) for m in model_types)
    if has_asset_model:
        production_model = to_production_model(config.production)
        if models.ProductionModel in model_types:
            all_models.append(production_model)
        if models.MarketModel in model_types or models.CogShop1Asset in model_types:
            market_model = to_market_asset_model(config.market, production_model.price_areas, market_name)
            settings = config.settings
            if production_model.root_asset.external_id is None:
                raise ValueError("The production model must have an external_id")
            market_model.set_root_asset(
                settings.shop_service_url,
                settings.organization_subdomain,
                settings.tenant_id,
                production_model.root_asset.external_id,
            )
            if models.MarketModel in model_types:
                all_models.append(market_model)
            if models.CogShop1Asset in model_types:
                cogshop_model = to_cogshop_asset_model(
                    config.cogshop,
                    production_model.watercourses,
                    config.settings.shop_version,
                    market_model.dayahead_processes,
                    market_model.rkom_processes,
                )
                all_models.append(cogshop_model)

    has_data_model = any(
        issubclass(m, (DataModel, models.PowerAssetModelDM)) and not issubclass(m, models.CogShop1Asset)
        for m in model_types
    )
    if has_data_model:
        if models.PowerAssetModelDM in model_types:
            power_asset_model = to_powerasset_model.to_asset_data_model(config.production)
            all_models.append(power_asset_model)

    return all_models



import datetime
import operator
import re
from functools import reduce
from typing import Any, Union

from cognite.client.utils._time import datetime_to_ms


def format_deep_diff_path(path: str) -> str:
    """
    Formats from deepdiff path to dot separated path
    ```
    >>> format_deep_diff_path(path="root['object_type']['object_name']['attribute']['name']")
    'object_type.object_name.attribute.name'

    ```
    """
    return re.sub(r"['\[\]]", "", path.replace("][", ".").replace("root", ""))


def get_dict_dot_keys(data_dict: dict, dot_keys: str) -> Union[int, str, datetime.datetime, Any]:
    """
    Get for arbitrarily nested keys that are dot separated.
    Handles that the yaml parser parses numeric keys as numbers
    and datetime keys as datetime.datetime(...)

    Solution is not complete,
    there might be other types that need to be handled

    Example:
    ```
    >>> get_dict_dot_keys(data_dict={"my_key": {"my_nested_key": 42,}}, dot_keys="my_key.my_nested_key")
    42

    ```
    """
    # TODO: This function should be simplified. It doesn't pass mypy checks, which have temporarily been disabled.

    keys = [int(k) if k.isdigit() else k for k in dot_keys.split(".")]
    while "datetime" in keys:
        i = keys.index("datetime")
        ts_str = str(keys.pop(i + 1))
        ts_str_tuple = ts_str.replace("datetime(", "").replace(")", "")
        dt_ts = datetime.datetime(*[int(x) for x in ts_str_tuple.split(",")])  # type: ignore[arg-type]
        keys[i] = dt_ts
    return reduce(operator.getitem, keys, data_dict)


def get_data_from_nested_dict(data_dict: dict, deep_diff_path: str) -> Union[int, str, datetime.datetime, Any]:
    dot_keys = format_deep_diff_path(deep_diff_path)
    return get_dict_dot_keys(data_dict, dot_keys)


def is_time_series_dict(data: Any) -> bool:
    return (
        isinstance(data, dict)
        and all(isinstance(k, datetime.datetime) for k in data.keys())
        and all(isinstance(v, (float, int)) for v in data.values())
    )


def str_datetime_to_ms(str_datetime: str, str_format: Union[str, None] = None) -> int:
    """
    Convert a string datetime to milliseconds since epoch.
    If no format is provided, the function will try to guess the format.

    >>> str_datetime_to_ms("2021-01-01 00:00:00")
    1609459200000
    >>> str_datetime_to_ms("2021-01-01 00:00")
    1609459200000
    >>> str_datetime_to_ms("2021-01-01")
    1609459200000
    >>> str_datetime_to_ms("2021-01-01 00:00:00", str_format="%Y-%m-%d %H:%M:%S")
    1609459200000

    """
    date_format = "%Y-%m-%d"
    date_time_formats = {
        0: date_format,
        1: f"{date_format} %H:%M",
        2: f"{date_format} %H:%M:%S",
    }

    if not str_format:
        str_format = date_time_formats[str_datetime.count(":")]
    datetime_ = datetime.datetime.strptime(str_datetime, str_format)
    if datetime_.tzinfo is None:
        datetime_ = datetime_.replace(tzinfo=datetime.timezone.utc)
    return datetime_to_ms(datetime_)



from pathlib import Path

from cognite.client.data_classes.data_modeling import DataModelId

from cognite.powerops.resync.models.base import PowerOpsGraphQLModel

graphql_file = Path(__file__).resolve().parent / "cogshop1.graphql"

cogshop1_graphql = PowerOpsGraphQLModel(
    name="CogShop",
    description="This is the first iteration of the cogshop model",
    graphql_file=graphql_file,
    id_=DataModelId("cogShop", "CogShop", "3"),
)

GRAPHQL_MODELS = {"cogshop1": cogshop1_graphql}



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from hashlib import sha256
from typing import Any, ClassVar, Optional

import pandas as pd
from cognite.client.data_classes import FileMetadata, Sequence
from cognite.client.data_classes._base import CogniteResource
from cognite.client.data_classes.data_modeling.ids import AbstractDataclass
from pydantic import BaseModel, ConfigDict, model_serializer, model_validator
from typing_extensions import Self

from cognite.powerops.utils.serialization import remove_read_only_fields


class CDFResource(BaseModel, ABC):
    content_key_hash: ClassVar[str] = "sha256_hash"
    model_config: ClassVar[ConfigDict] = ConfigDict(arbitrary_types_allowed=True)

    @model_serializer
    def ser_model(self) -> dict[str, Any]:
        return self.dump(camel_case=False)

    def dump(self, camel_case: bool = False) -> dict[str, Any]:
        return remove_read_only_fields(self._dump(camel_case))

    @abstractmethod
    def _dump(self, camel_case: bool = False) -> dict[str, Any]:
        raise NotImplementedError()

    @property
    def cdf_resource(self) -> CogniteResource:
        raise NotImplementedError

    def __repr__(self) -> str:
        return f"{type(self).__name}({self.external_id=})"

    def __str__(self) -> str:
        return repr(self)


class CDFSequence(CDFResource):
    sequence: Sequence
    content: Optional[pd.DataFrame] = None

    @property
    def cdf_resource(self) -> Sequence:
        return self.sequence

    @model_validator(mode="before")
    @classmethod
    def parse_dict(cls, value):
        if isinstance(value, dict) and "sequence" not in value:
            return {"sequence": Sequence._load(value)}
        return value

    @property
    def external_id(self):
        return self.sequence.external_id

    def dump(self, camel_case: bool = False) -> dict[str, Any]:
        output = super().dump(camel_case)
        if (columns := output.get("columns")) and isinstance(columns, list):
            for no, column in enumerate(columns):
                output["columns"][no] = remove_read_only_fields(column)
        return output

    @classmethod
    def calculate_hash(cls, content: pd.DataFrame) -> str:
        return sha256(pd.util.hash_pandas_object(content, index=True).values, usedforsecurity=False).hexdigest()

    def _dump(self, camel_case: bool = False) -> dict[str, Any]:
        if self.content is not None:
            sha256_hash = self.calculate_hash(self.content)
        elif self.sequence.metadata and self.sequence.metadata.get(self.content_key_hash) is not None:
            sha256_hash = self.sequence.metadata[self.content_key_hash]
        else:
            sha256_hash = "Missing Content"

        if self.sequence.metadata:
            self.sequence.metadata[self.content_key_hash] = sha256_hash
        else:
            self.sequence.metadata = {self.content_key_hash: sha256_hash}
        return self.sequence.dump(camel_case=camel_case)

    @classmethod
    def _load(cls, data: dict[str, Any]) -> Self:
        return cls(sequence=Sequence._load(data))


class CDFFile(CDFResource):
    meta: FileMetadata
    content: Optional[bytes] = None

    @property
    def cdf_resource(self) -> FileMetadata:
        return self.meta

    @model_validator(mode="before")
    @classmethod
    def parse_dict(cls, value):
        if isinstance(value, dict) and "meta" not in value:
            return {"meta": FileMetadata._load(value)}
        return value

    @property
    def external_id(self):
        return self.meta.external_id

    @classmethod
    def calculate_hash(cls, content: bytes) -> str:
        # The replacement is used to ensure that the hash is the same on Windows and Linux
        return sha256(content.replace(b"\r\n", b"\n"), usedforsecurity=False).hexdigest()

    def _dump(self, camel_case: bool = False) -> dict[str, Any]:
        if self.content is None and (self.meta.metadata or {}).get(self.content_key_hash) is None:
            sha256_hash = "Missing Content"
        elif self.content is not None:
            sha256_hash = self.calculate_hash(self.content)
        else:
            sha256_hash = self.meta.metadata[self.content_key_hash]
        if self.meta.metadata:
            self.meta.metadata[self.content_key_hash] = sha256_hash
        else:
            self.meta.metadata = {self.content_key_hash: sha256_hash}
        return self.meta.dump(camel_case=camel_case)

    @classmethod
    def _load(cls, data: dict[str, Any]) -> Self:
        return cls(meta=FileMetadata._load(data))


# The extra class is because the AbstractDataClass does allow it
# subclasses to be instantiated. This is a private implementation detail in the Python-SDK, but it is useful
# for us here in the adapter methods of the CDF API.
@dataclass(frozen=True)
class _Dummy(AbstractDataclass): ...


@dataclass(frozen=True)
class SpaceId(_Dummy):
    space: str

    @property
    def external_id(self) -> str:
        return self.space

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {"space": self.space}



from __future__ import annotations

import warnings
from typing import Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from cognite.client.data_classes import TimeSeries as CogniteTimeSeries
from pydantic import Field
from pydantic import field_validator, model_validator

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    GraphQLCore,
    ResourcesWrite,
    TimeSeries,
)


__all__ = ["PriceArea", "PriceAreaList", "PriceAreaFields", "PriceAreaTextFields"]


PriceAreaTextFields = Literal[
    "name",
    "capacity_price_up",
    "capacity_price_down",
    "activation_price_up",
    "activation_price_down",
    "relative_activation",
    "total_capacity_allocation_up",
    "total_capacity_allocation_down",
    "own_capacity_allocation_up",
    "own_capacity_allocation_down",
]
PriceAreaFields = Literal[
    "name",
    "capacity_price_up",
    "capacity_price_down",
    "activation_price_up",
    "activation_price_down",
    "relative_activation",
    "total_capacity_allocation_up",
    "total_capacity_allocation_down",
    "own_capacity_allocation_up",
    "own_capacity_allocation_down",
]

_PRICEAREA_PROPERTIES_BY_FIELD = {
    "name": "name",
    "capacity_price_up": "capacityPriceUp",
    "capacity_price_down": "capacityPriceDown",
    "activation_price_up": "activationPriceUp",
    "activation_price_down": "activationPriceDown",
    "relative_activation": "relativeActivation",
    "total_capacity_allocation_up": "totalCapacityAllocationUp",
    "total_capacity_allocation_down": "totalCapacityAllocationDown",
    "own_capacity_allocation_up": "ownCapacityAllocationUp",
    "own_capacity_allocation_down": "ownCapacityAllocationDown",
}


class PriceAreaGraphQL(GraphQLCore):
    """This represents the reading version of price area, used
    when data is retrieved from CDF using GraphQL.

    It is used when retrieving data from CDF using GraphQL.

    Args:
        space: The space where the node is located.
        external_id: The external id of the price area.
        data_record: The data record of the price area node.
        name: Name for the PriceArea.
        capacity_price_up: The capacity price up field.
        capacity_price_down: The capacity price down field.
        activation_price_up: The mFRR activation price (TBC)
        activation_price_down: The mFRR activate price (TBC)
        relative_activation: Value between -1 (100 % activation down) and 1 (100 % activation down)
        total_capacity_allocation_up: The total capacity allocation up field.
        total_capacity_allocation_down: The total capacity allocation down field.
        own_capacity_allocation_up: The own capacity allocation up field.
        own_capacity_allocation_down: The own capacity allocation down field.
    """

    view_id = dm.ViewId("power-ops-afrr-bid", "PriceArea", "1")
    name: Optional[str] = None
    capacity_price_up: Union[TimeSeries, dict, None] = Field(None, alias="capacityPriceUp")
    capacity_price_down: Union[TimeSeries, dict, None] = Field(None, alias="capacityPriceDown")
    activation_price_up: Union[TimeSeries, dict, None] = Field(None, alias="activationPriceUp")
    activation_price_down: Union[TimeSeries, dict, None] = Field(None, alias="activationPriceDown")
    relative_activation: Union[TimeSeries, dict, None] = Field(None, alias="relativeActivation")
    total_capacity_allocation_up: Union[TimeSeries, dict, None] = Field(None, alias="totalCapacityAllocationUp")
    total_capacity_allocation_down: Union[TimeSeries, dict, None] = Field(None, alias="totalCapacityAllocationDown")
    own_capacity_allocation_up: Union[TimeSeries, dict, None] = Field(None, alias="ownCapacityAllocationUp")
    own_capacity_allocation_down: Union[TimeSeries, dict, None] = Field(None, alias="ownCapacityAllocationDown")

    @model_validator(mode="before")
    def parse_data_record(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values
        if "lastUpdatedTime" in values or "createdTime" in values:
            values["dataRecord"] = DataRecordGraphQL(
                created_time=values.pop("createdTime", None),
                last_updated_time=values.pop("lastUpdatedTime", None),
            )
        return values

    def as_read(self) -> PriceArea:
        """Convert this GraphQL format of price area to the reading format."""
        if self.data_record is None:
            raise ValueError("This object cannot be converted to a read format because it lacks a data record.")
        return PriceArea(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecord(
                version=0,
                last_updated_time=self.data_record.last_updated_time,
                created_time=self.data_record.created_time,
            ),
            name=self.name,
            capacity_price_up=self.capacity_price_up,
            capacity_price_down=self.capacity_price_down,
            activation_price_up=self.activation_price_up,
            activation_price_down=self.activation_price_down,
            relative_activation=self.relative_activation,
            total_capacity_allocation_up=self.total_capacity_allocation_up,
            total_capacity_allocation_down=self.total_capacity_allocation_down,
            own_capacity_allocation_up=self.own_capacity_allocation_up,
            own_capacity_allocation_down=self.own_capacity_allocation_down,
        )


class PriceArea(DomainModel):
    """This represents the reading version of price area.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the price area.
        data_record: The data record of the price area node.
        name: Name for the PriceArea.
        capacity_price_up: The capacity price up field.
        capacity_price_down: The capacity price down field.
        activation_price_up: The mFRR activation price (TBC)
        activation_price_down: The mFRR activate price (TBC)
        relative_activation: Value between -1 (100 % activation down) and 1 (100 % activation down)
        total_capacity_allocation_up: The total capacity allocation up field.
        total_capacity_allocation_down: The total capacity allocation down field.
        own_capacity_allocation_up: The own capacity allocation up field.
        own_capacity_allocation_down: The own capacity allocation down field.
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("power-ops-types", "PriceArea")
    name: str
    capacity_price_up: Union[TimeSeries, str, None] = Field(None, alias="capacityPriceUp")
    capacity_price_down: Union[TimeSeries, str, None] = Field(None, alias="capacityPriceDown")
    activation_price_up: Union[TimeSeries, str, None] = Field(None, alias="activationPriceUp")
    activation_price_down: Union[TimeSeries, str, None] = Field(None, alias="activationPriceDown")
    relative_activation: Union[TimeSeries, str, None] = Field(None, alias="relativeActivation")
    total_capacity_allocation_up: Union[TimeSeries, str, None] = Field(None, alias="totalCapacityAllocationUp")
    total_capacity_allocation_down: Union[TimeSeries, str, None] = Field(None, alias="totalCapacityAllocationDown")
    own_capacity_allocation_up: Union[TimeSeries, str, None] = Field(None, alias="ownCapacityAllocationUp")
    own_capacity_allocation_down: Union[TimeSeries, str, None] = Field(None, alias="ownCapacityAllocationDown")


class PriceAreaList(DomainModelList[PriceArea]):
    """List of price areas in the read version."""

    _INSTANCE = PriceArea


def _create_price_area_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None
